Package: GenProSeq
Type: Package
Title: Generating Protein Sequences with Deep Generative Models
Description: Generative modeling for protein engineering is key to solving fundamental problems in synthetic biology, medicine, and material science. Machine learning has enabled us to generate useful protein sequences on a variety of scales. Generative models are machine learning methods which seek to model the distribution underlying the data, allowing for the generation of novel samples with similar properties to those on which the model was trained. Generative models of proteins can learn biologically meaningful representations helpful for a variety of downstream tasks. Furthermore, they can learn to generate protein sequences that have not been observed before and to assign higher probability to protein sequences that satisfy desired criteria. In this package, common deep generative models for protein sequences, such as variational autoencoder (VAE), generative adversarial networks (GAN), and autoregressive models are available. In the VAE and GAN, the Word2vec is used for embedding. The transformer encoder is applied to protein sequences for the autoregressive model.
Version: 1.16.0
Date: 2024-02-06
Authors@R: c(person(given="Dongmin", family="Jung", email="dmdmjung@gmail.com", role=c("cre", "aut"), comment = c(ORCID = "0000-0001-7499-8422")))
LazyData: FALSE
Depends: keras, mclust, R (>= 4.2)
Imports: tensorflow, word2vec, DeepPINCS, ttgsea, CatEncoders,
        reticulate, stats
Suggests: VAExprs, stringdist, knitr, testthat, rmarkdown
License: Artistic-2.0
biocViews: Software, Proteomics
NeedsCompilation: no
VignetteBuilder: knitr
git_url: https://git.bioconductor.org/packages/GenProSeq
git_branch: RELEASE_3_23
git_last_commit: e5100b8
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
Packaged: 2026-04-28 22:15:58 UTC; biocbuild
Author: Dongmin Jung [cre, aut] (ORCID:
    <https://orcid.org/0000-0001-7499-8422>)
Maintainer: Dongmin Jung <dmdmjung@gmail.com>
Built: R 4.6.0; ; 2026-04-29 04:44:38 UTC; unix

export("fit_VAE")
export("gen_VAE")
export("fit_GAN")
export("gen_GAN")
export("fit_ART")
export("gen_ART")
export("layer_embedding_token_position")
export("layer_transformer_encoder")
export("prot2vec")
export("vec2prot")
export("prot_seq_check")
import("keras")
import("mclust", except = c("map"))
importFrom("word2vec", "word2vec", "word2vec_similarity")
importFrom("CatEncoders", "LabelEncoder.fit", "transform", "inverse.transform")
importFrom("DeepPINCS", "multiple_sampling_generator", "seq_preprocessing", "get_seq_encode_pad")
importFrom("ttgsea", "sampling_generator")
importFrom("stats", "embed", "predict")

# GenProSeq

## Introduction

Generative modeling for protein engineering is key to solving fundamental problems in synthetic biology, medicine, and material science. Machine learning has enabled us to generate useful protein sequences on a variety of scales. Generative models are machine learning methods which seek to model the distribution underlying the data, allowing for the generation of novel samples with similar properties to those on which the model was trained. Generative models of proteins can learn biologically meaningful representations helpful for a variety of downstream tasks. Furthermore, they can learn to generate protein sequences that have not been observed before and to assign higher probability to protein sequences that satisfy desired criteria. In this package, common deep generative models for protein sequences, such as variational autoencoder (VAE), generative adversarial networks (GAN), and autoregressive models are available. In the VAE and GAN, the Word2vec is used for embedding. The transformer encoder is applied to protein sequences for the autoregressive model.

The first step in molecular machine learning is to convert the molecular data into a numerical format suitable for the machine learning models. As an example of a raw representation is the sparse or one-hot encoding. Here, each amino acid is encoded as a vector, where a bit is set according to an index using the known amino acids. The naive way of representing a word in vector form is one hot representation but it is a very ineffective way for representing a huge corpus. In a more effective way, we need some semantic similarities to nearby points, thus creating the representation bring beneficial information about the word actual meaning, called word embedding models. In natural language processing, often large, one-hot encoded vectors are transformed into smaller dimension vectors, by finding word relations based on proximity or positions of similar words using the contexts of nearby neighbors. The Word2vec is one of the most common word embedding models that gives low-dimensional dense vector (word embedding) of the term (words and phrases). The Word2vec has also been applied to amino acids as a basis for traditional machine learning methods.

The generative adversarial network (GAN) is able to mimic any distribution of data in any domain containing images, music, speech, and prose. The GAN is an example of a network that use unsupervised learning to train two models in parallel. The network is forced to efficiently represent the training data, making it more effective at generating data similar to the training data. The GAN is made up of a discriminator and a generator that compete in a two-player minimax game. The generator generates a simulated example given as its input from a specified distribution. The objective of the generator is to produce an output that is so close to real that it confuses the discriminator in being able to differentiate the fake data from the real data. Both the generated examples and authentic ones are fed to the discriminator network, whose job is to distinguish between the fake and the real data. Generative adversarial nets can be extended to a conditional model if both the generator and discriminator are conditioned on some extra information, such as class labels or data from other views. The conditional GAN (CGAN) is based on vanilla GAN with additional conditional input to generator and discriminator. This extracts features based on the modelling of the conditional input. The auxiliary classifier GAN (ACGAN) is an extension of CGAN that adds conditional input only to the generator.

The autoencoder is the unsupervised neural network approach to learn how to encode data efficiently in a latent space. In the autoencoder, an encoder maps data from the input space to the latent space and a decoder is used to reconstruct the input data from the encoded latent data. The variational autoencoder (VAE) is a class of autoencoder where the encoder module is used to learn the parameter of a distribution and the decoder is used to generate examples from samples drawn from the learned distribution. The VAE does not allow us to constrain the generated sample to have a particular characteristic, but one may want to draw samples with the desired feature. A question then arises on how to endow the model to create targeted samples rather than completely random ones. The conditional variational autoencoder (CVAE) is designed to generate desired samples by including additional conditioning information.

Language models learn the syntactic order of a language from a given set of examples as probability values to predict a sequence of words. A language model is an indispensable ingredient in many advanced natural language processing (NLP) tasks such as text summarization, machine translation, and language generation. The language modeling is a typical self-supervised objective, for it does not require any labels. Language models developed for NLP tasks have evolved from statistical language models to neural language models. Approaches such as n-gram and hidden Markov models have become the primary ingredient in developing statistical language models. In contrast, various deep learning neural networks build neural language models. In the autoregressive (AR) language modeling, models are tasked with generating subsequent tokens based on previously generated tokens. Thus the autoregressive generative model predicts the next amino acid in a protein given the amino acid sequence up to that point. The autoregressive model generates proteins one amino acid at a time. For one step of generation, it takes a context sequence of amino acids as input and outputs a probability distribution over amino acids. We sample from that distribution and then update the context sequence with the sampled amino acid. This process repeats until a protein of desired length has been generated.

The Transformer architecture demonstrates impressive text generation capabilities. This perspective is adapted to protein engineering by training the autoregressive language model with the Transformer encoder on amino acid sequences. Compared to conventional sequential models such as RNN, the Transformer is a new architecture that is more effective in modeling long-term dependence in temporal sequences, where all tokens will be equally considered during the attention operation. And it is more efficient in training while eliminating sequential dependencies from previous tokens. The Transformer models can generally overcome the inherent limitations of classic neural network architectures. For example, these models overcome the problem of speed inherent in RNN, LSTM, or GRU, which require sequential operations that are slow in nature. They can also overcome the long-term dependencies problem of CNN, which can never accurately handle long-range dependences in the text corpus. Unlike RNN, the Transformer cannot consider the order of the input data, by using not only tokens but also positions for embedding. Since the Transformer has a very powerful ability to model sequential data, it becomes the most popular backbone of NLP applications. The Transformer architecture is a nonrecurrent architecture with a series of attention-based blocks. Each block is composed of a multi-head attention layer and a position-wise feedforward layer with an add and normalize layer in between. These layers process input sequences simultaneously, in parallel, independently of sequential order.





## Installation

For this package, it is required to install tensorflow and kears in python and connect to R (https://tensorflow.rstudio.com, https://keras.rstudio.com). This package is currently available for tensorflow 2.8.0, keras 2.8.0, and python 3.9.9. However, it will be maintained to keep up with changes.
To install the development version from GitHub:
``` 
devtools::install_github("dongminjung/GenProSeq")
```





## Example

## GAN

The sequences of the PTEN are used to train the GAN. The input sequences must be aligned to have the same length. To train the model, we can use the function "fit_GAN" with aligned sequence data. To generate sequences, we can use the function "gen_GAN" with the trained model.
It is expected that the model can rapidly generate highly diverse novel functional proteins within the allowed biological constraints of the sequence space. Note that the same dataset is used for training and validation here.
```
library(GenProSeq)

# model parameters
length_seq <- 403
embedding_dim <- 8
latent_dim <- 4
epochs <- 20
batch_size <- 64

# GAN
GAN_result <- fit_GAN(prot_seq = example_PTEN,
                      length_seq = length_seq,
                      embedding_dim = embedding_dim,
                      latent_dim = latent_dim,
                      intermediate_generator_layers = list(
                          layer_dense(units = 16),
                          layer_dense(units = 128)),
                      intermediate_discriminator_layers = list(
                          layer_dense(units = 128, activation = "relu"),
                          layer_dense(units = 16, activation = "relu")),
                      prot_seq_val = example_PTEN,
                      epochs = epochs,
                      batch_size = batch_size)
```

The model architecture of the generator is shown below.
```
ttgsea::plot_model(GAN_result$generator)
```

![history plot](docs/GAN_architecture_generator.jpeg)

The model architecture of the discriminator is shown below.
```
ttgsea::plot_model(GAN_result$discriminator)
```

![history plot](docs/GAN_architecture_discriminator.jpeg)

In a graphical manner, a sequence logo shows the probability of occurrence for each symbol at specific positions. Thus, the height of a symbol indicates the relative frequency of the symbol at that position. The sequence logo of the first 20 amino acids of the generated protein sequences is shown below.
```
set.seed(1)
gen_prot_GAN <- gen_GAN(GAN_result, num_seq = 100)
ggseqlogo::ggseqlogo(substr(gen_prot_GAN$gen_seq, 1, 20))
```

![history plot](docs/GAN_seqlogo_gen.png)

The sequence logo of the first 20 amino acids of the real protein sequences is shown below. Here, it is assumed that the conserved position is the first amino acid. The sequence logo is a visualization approach to graphically represent the sequence conservation.
```
ggseqlogo::ggseqlogo(substr(example_PTEN, 1, 20))
```

![history plot](docs/GAN_seqlogo_real.png)



## VAE

Consider the aligned sequences of the luxA for traing the CVAE with labels. Suppose that the label is the third amino acid of each sequence. Thus there are two class labels. Using the function "fit_VAE", we build an encoder model that takes a protein sequence and projects it on the latent space and a decoder model that goes from the latent space back to the amino acid representation. And then, the function "gen_VAE" generates sequences with the desired labels. Note that the same dataset is used for training and validation.
```
library(GenProSeq)
label <- substr(example_luxA, 3, 3)

# model parameters
length_seq <- 360
embedding_dim <- 8
batch_size <- 128
epochs <- 20

# CVAE
VAE_result <- fit_VAE(prot_seq = example_luxA,
                    label = label,
                    length_seq = length_seq,
                    embedding_dim = embedding_dim,
                    embedding_args = list(iter = 20),
                    intermediate_encoder_layers = list(layer_dense(units = 128),
                                                        layer_dense(units = 16)),
                    intermediate_decoder_layers = list(layer_dense(units = 16),
                                                        layer_dense(units = 128)),
                    prot_seq_val = example_luxA,
                    label_val = label,
                    epochs = epochs,
                    batch_size = batch_size,
                    use_generator = FALSE,
                    optimizer = keras::optimizer_adam(clipnorm = 0.1),
                    callbacks = keras::callback_early_stopping(
                        monitor = "val_loss",
                        patience = 10,
                        restore_best_weights = TRUE))
```

The plot for model architecture of the CVAE is drawn below.
```
VAExprs::plot_vae(VAE_result$model)
```

![history plot](docs/VAE_architecture.jpeg)

The sequence logo of the first 20 amino acids of the generated protein sequences with the label "I" is shown below.
```
set.seed(1)
gen_prot_VAE <- gen_VAE(VAE_result, label = rep("I", 100), num_seq = 100)
ggseqlogo::ggseqlogo(substr(gen_prot_VAE$gen_seq, 1, 20))
```

![history plot](docs/VAE_seqlogo_genI.png)

The sequence logo of the first 20 amino acids of the generated protein sequences with the label "L" is shown below.
```
gen_prot_VAE <- gen_VAE(VAE_result, label = rep("L", 100), num_seq = 100)
ggseqlogo::ggseqlogo(substr(gen_prot_VAE$gen_seq, 1, 20))
```

![history plot](docs/VAE_seqlogo_genL.png)

The sequence logo of the first 20 amino acids of the real protein sequences is shown below.
```
ggseqlogo::ggseqlogo(substr(example_luxA, 1, 20))
```

![history plot](docs/VAE_seqlogo_real.png)



## AR with Transformer

The SARS coronavirus 3C-like Protease is used to train the autoregressive language model with the Transformer. The same dataset is used for training and validation.
```
library(GenProSeq)
prot_seq <- DeepPINCS::SARS_CoV2_3CL_Protease

# model parameters
length_seq <- 10
embedding_dim <- 16
num_heads <- 2
ff_dim <- 16
num_transformer_blocks <- 2
batch_size <- 32
epochs <- 500

# ART
ART_result <- fit_ART(prot_seq = prot_seq,
                    length_seq = length_seq,
                    embedding_dim = embedding_dim,
                    num_heads = num_heads,
                    ff_dim = ff_dim,
                    num_transformer_blocks = num_transformer_blocks,
                    layers = list(layer_dropout(rate = 0.1),
                                layer_dense(units = 32, activation = "relu"),
                                layer_dropout(rate = 0.1)),
                    prot_seq_val = prot_seq,
                    epochs = epochs,
                    batch_size = batch_size,
                    use_generator = TRUE,
                    callbacks = callback_early_stopping(
                        monitor = "val_loss",
                        patience = 50,
                        restore_best_weights = TRUE))
```

The model architecture is shown below.
```
ttgsea::plot_model(ART_result$model)
```

![history plot](docs/ART_architecture.jpeg)

The learned autoregressive model generates new sequences of proteins. After the language model generates a conditional probability distribution over vocabulary for the given input sequence, we need to decide how to choose the next word or token from the distribution. The greedy search simply selects the token with the highest probability value as its next token. Rather than just considering the highest probable word, the top b words is taken as the next word in the beam search. Here b is also called the beam size which is itself a parameter. Now to compute the second word each of these b first words is fed and top b words are obtained again. The process continues till the end. In temperature sampling, we sample the token from the modified conditional probability distribution over the vocabulary for the given temperature value. In the top-k sampling, the k most likely next tokens are filtered and their probabilities are adjusted among only those k tokens. In a similar manner, the top-p sampling is another way to exclude very low probability tokens and it finds the smallest set of tokens that have summed probability at least p.
```
seed_prot <- "SGFRKMAFPS"
gen_ART(ART_result, seed_prot, length_AA = 20, method = "greedy")
"SGFRKMAFPSGKVEGCMVQVTCGTTTLNGL"
substr(prot_seq, 1, 30)
"SGFRKMAFPSGKVEGCMVQVTCGTTTLNGL"
gen_ART(ART_result, seed_prot, length_AA = 20, method = "beam", b = 5)
"SGFRKMAFPSGKVEGCMVQVTCGTTTLNGL"
substr(prot_seq, 1, 30)
"SGFRKMAFPSGKVEGCMVQVTCGTTTLNGL"
gen_ART(ART_result, seed_prot, length_AA = 20, method = "temperature", t = 0.1)
"SGFRKMAFPSGKVEGCMVQVTCGTTTLNGL"
substr(prot_seq, 1, 30)
"SGFRKMAFPSGKVEGCMVQVTCGTTTLNGL"
gen_ART(ART_result, seed_prot, length_AA = 20, method = "top_k", k = 3)
"SGFRKMAFPSGKVEGCMVQVTCGTTTLNGL"
substr(prot_seq, 1, 30)
"SGFRKMAFPSGKVEGCMVQVTCGTTTLNGL"
gen_ART(ART_result, seed_prot, length_AA = 20, method = "top_p", p = 0.75)
"SGFRKMAFPSGVVELDMVPMVRVVVRHVHV"
substr(prot_seq, 1, 30)
"SGFRKMAFPSGKVEGCMVQVTCGTTTLNGL"
```

We can compute pairwise similarities between the real and generated protein sequences. The function "stringsim" vary between 0 for strings that are not similar at all, to 1 for strings that are identical. One advantage of the string similarity over the string distance function is that similarities are easier to interpret because they are normalized.
```
stringdist::stringsim(gen_ART(ART_result, seed_prot, length_AA = 20, method = "greedy"),
                    substr(prot_seq, 1, 30))
1
stringdist::stringsim(gen_ART(ART_result, seed_prot, length_AA = 30, method = "greedy"),
                    substr(prot_seq, 1, 40))
1
stringdist::stringsim(gen_ART(ART_result, seed_prot, length_AA = 40, method = "greedy"),
                    substr(prot_seq, 1, 50))
1
stringdist::stringsim(gen_ART(ART_result, seed_prot, length_AA = 50, method = "greedy"),
                    substr(prot_seq, 1, 60))
1
```





# References

Barbosa, V. A. F., Santana, M. A., Andrade, M. K. S., Lima, R. C. F., & Santos, W. P. (2020). Deep Learning for Data Analytics: Foundations, Biomedical Applications, and Challenges. Academic Press.

Cinelli, L. P., Marins, M. A., da Silva, E. A. B., & Netto, S. L. (2021). Variational Methods for Machine Learning with Applications to Deep Networks. Springer.

Dash, S., Acharya, B. R., Mittal, M., Abraham, A., & Kelemen, A. (Eds.). (2020). Deep learning techniques for biomedical and health informatics. Springer.

Deepak, P., Chakraborty, T., & Long, C. (2021). Data Science for Fake News: Surveys and Perspectives. Springer.

Dong, G., & Pei, J. (2007). Sequence data mining. Springer.

Frazer, J., Notin, P., Dias, M., Gomez, A., Brock, K., Gal, Y., & Marks, D. (2020). Large-scale clinical interpretation of genetic variants using evolutionary data and deep learning. bioRxiv.

Gagniuc, P. A. (2021). Algorithms in Bioinformatics: Theory and Implementation. Wiley & Sons.

Hawkins-Hooker, A., Depardieu, F., Baur, S., Couairon, G., Chen, A., & Bikard, D. (2020). Generating functional protein variants with variational autoencoders. bioRxiv.

Hemanth, J., Bhatia, M., & Geman, O. (2020). Data Visualization and Knowledge Engineering: Spotting Data Points with Artificial Intelligence. Springer.

Lappin, S. (2021). Deep learning and linguistic representation. CRC Press.

Liebowitz, J. (Ed.). (2020). Data Analytics and AI. CRC Press.

Liu, Z., Lin, Y., & Sun, M. (2020). Representation learning for natural language processing. Springer.

Madani, A., McCann, B., Naik, N., Keskar, N. S., Anand, N., Eguchi, R. R., Huang, P., & Socher, R. (2020). Progen: Language modeling for protein generation. arXiv:2004.03497.

Pearson, R. K. (2018). Exploratory data analysis using R. CRC Press.

Pedrycz, W., & Chen, S. M. (Eds.). (2020). Deep Learning: Concepts and Architectures. Springer.

Peter, J. D., Fernandes, S. L., Thomaz, C. E., & Viriri, S. (Eds.). (2019). Computer aided intervention and diagnostics in clinical and medical images. Springer.

Repecka, D., et al. (2019). Expanding functional protein sequence space using generative adversarial networks. bioRxiv.

Suguna, S. K., Dhivya, M., & Paiva, S. (Eds.). (2021). Artificial Intelligence (AI): Recent Trends and Applications. CRC Press.

Sun, S., Mao, L., Dong, Z., & Wu, L. (2019). Multiview machine learning. Springer.

Wolkenhauer, O. (2020). Systems Medicine: Integrative, Qualitative and Computational Approaches. Academic Press.

Wu, Z., Johnston, K. E., Arnold, F. H., & Yang, K. K. (2021). Protein sequence design with deep generative models. arXiv:2104.04457.

## ----eval=TRUE----------------------------------------------------------------
if (keras::is_keras_available() & reticulate::py_available()) {
    library(GenProSeq)
    data("example_PTEN")
    
    # model parameters
    length_seq <- 403
    embedding_dim <- 8
    latent_dim <- 4
    epochs <- 20
    batch_size <- 64
    
    # GAN
    GAN_result <- fit_GAN(prot_seq = example_PTEN,
                        length_seq = length_seq,
                        embedding_dim = embedding_dim,
                        latent_dim = latent_dim,
                        intermediate_generator_layers = list(
                            layer_dense(units = 16),
                            layer_dense(units = 128)),
                        intermediate_discriminator_layers = list(
                            layer_dense(units = 128, activation = "relu"),
                            layer_dense(units = 16, activation = "relu")),
                        prot_seq_val = example_PTEN,
                        epochs = epochs,
                        batch_size = batch_size)
}

## ----eval=TRUE----------------------------------------------------------------
if (keras::is_keras_available() & reticulate::py_available()) {
    ttgsea::plot_model(GAN_result$generator)
}

## ----eval=TRUE----------------------------------------------------------------
if (keras::is_keras_available() & reticulate::py_available()) {
    ttgsea::plot_model(GAN_result$discriminator)
}

## ----eval=TRUE----------------------------------------------------------------
if (keras::is_keras_available() & reticulate::py_available()) {
    set.seed(1)
    gen_prot_GAN <- gen_GAN(GAN_result, num_seq = 100)
    if (require(ggseqlogo)) {
        ggseqlogo::ggseqlogo(substr(gen_prot_GAN$gen_seq, 1, 20))
    }
}

## ----eval=TRUE----------------------------------------------------------------
if (keras::is_keras_available() & reticulate::py_available()) {
    if (require(ggseqlogo)) {
        ggseqlogo::ggseqlogo(substr(example_PTEN, 1, 20))
    }
}

## ----eval=TRUE----------------------------------------------------------------
if (keras::is_keras_available() & reticulate::py_available()) {
    library(GenProSeq)
    data("example_luxA")
    label <- substr(example_luxA, 3, 3)
    
    # model parameters
    length_seq <- 360
    embedding_dim <- 8
    batch_size <- 128
    epochs <- 20
    
    # CVAE
    VAE_result <- fit_VAE(prot_seq = example_luxA,
                        label = label,
                        length_seq = length_seq,
                        embedding_dim = embedding_dim,
                        embedding_args = list(iter = 20),
                        intermediate_encoder_layers = list(layer_dense(units = 128),
                                                            layer_dense(units = 16)),
                        intermediate_decoder_layers = list(layer_dense(units = 16),
                                                            layer_dense(units = 128)),
                        prot_seq_val = example_luxA,
                        label_val = label,
                        epochs = epochs,
                        batch_size = batch_size,
                        use_generator = FALSE,
                        optimizer = keras::optimizer_adam(clipnorm = 0.1),
                        callbacks = keras::callback_early_stopping(
                            monitor = "val_loss",
                            patience = 10,
                            restore_best_weights = TRUE))
}

## ----eval=TRUE----------------------------------------------------------------
if (keras::is_keras_available() & reticulate::py_available()) {
    VAExprs::plot_vae(VAE_result$model)
}

## ----eval=TRUE----------------------------------------------------------------
if (keras::is_keras_available() & reticulate::py_available()) {
    set.seed(1)
    gen_prot_VAE <- gen_VAE(VAE_result, label = rep("I", 100), num_seq = 100)
    if (require(ggseqlogo)) {
        ggseqlogo::ggseqlogo(substr(gen_prot_VAE$gen_seq, 1, 20))
    }
}

## ----eval=TRUE----------------------------------------------------------------
if (keras::is_keras_available() & reticulate::py_available()) {
    gen_prot_VAE <- gen_VAE(VAE_result, label = rep("L", 100), num_seq = 100)
    if (require(ggseqlogo)) {
        ggseqlogo::ggseqlogo(substr(gen_prot_VAE$gen_seq, 1, 20))
    }
}

## ----eval=TRUE----------------------------------------------------------------
if (keras::is_keras_available() & reticulate::py_available()) {
    if (require(ggseqlogo)) {
        ggseqlogo::ggseqlogo(substr(example_luxA, 1, 20))
    }
}

## ----eval=TRUE----------------------------------------------------------------
if (keras::is_keras_available() & reticulate::py_available()) {
    library(GenProSeq)
    prot_seq <- DeepPINCS::SARS_CoV2_3CL_Protease
    
    # model parameters
    length_seq <- 10
    embedding_dim <- 16
    num_heads <- 2
    ff_dim <- 16
    num_transformer_blocks <- 2
    batch_size <- 32
    epochs <- 100
    
    # ART
    ART_result <- fit_ART(prot_seq = prot_seq,
                        length_seq = length_seq,
                        embedding_dim = embedding_dim,
                        num_heads = num_heads,
                        ff_dim = ff_dim,
                        num_transformer_blocks = num_transformer_blocks,
                        layers = list(layer_dropout(rate = 0.1),
                                    layer_dense(units = 32, activation = "relu"),
                                    layer_dropout(rate = 0.1)),
                        prot_seq_val = prot_seq,
                        epochs = epochs,
                        batch_size = batch_size,
                        use_generator = FALSE,
                        callbacks = callback_early_stopping(
                            monitor = "val_loss",
                            patience = 50,
                            restore_best_weights = TRUE))
}

## ----eval=TRUE----------------------------------------------------------------
if (keras::is_keras_available() & reticulate::py_available()) {
    ttgsea::plot_model(ART_result$model)
}

## ----eval=TRUE----------------------------------------------------------------
if (keras::is_keras_available() & reticulate::py_available()) {
    set.seed(1)
    seed_prot <- "SGFRKMAFPS"
    print(gen_ART(ART_result, seed_prot, length_AA = 20, method = "greedy"))
    print(substr(prot_seq, 1, 30))
    print(gen_ART(ART_result, seed_prot, length_AA = 20, method = "beam", b = 5))
    print(substr(prot_seq, 1, 30))
    print(gen_ART(ART_result, seed_prot, length_AA = 20, method = "temperature", t = 0.1))
    print(substr(prot_seq, 1, 30))
    print(gen_ART(ART_result, seed_prot, length_AA = 20, method = "top_k", k = 3))
    print(substr(prot_seq, 1, 30))
    print(gen_ART(ART_result, seed_prot, length_AA = 20, method = "top_p", p = 0.75))
    print(substr(prot_seq, 1, 30))
}

## ----eval=TRUE----------------------------------------------------------------
if (keras::is_keras_available() & reticulate::py_available()) {
    print(stringdist::stringsim(gen_ART(ART_result, seed_prot, length_AA = 20, method = "greedy"),
                        substr(prot_seq, 1, 30)))
    print(stringdist::stringsim(gen_ART(ART_result, seed_prot, length_AA = 30, method = "greedy"),
                        substr(prot_seq, 1, 40)))
    print(stringdist::stringsim(gen_ART(ART_result, seed_prot, length_AA = 40, method = "greedy"),
                        substr(prot_seq, 1, 50)))
    print(stringdist::stringsim(gen_ART(ART_result, seed_prot, length_AA = 50, method = "greedy"),
                        substr(prot_seq, 1, 60)))
}

## ----eval=TRUE----------------------------------------------------------------
sessionInfo()

prot_seq_check <- function(prot_seq, label = NULL) {
    result <- NULL
    
    if (!is.null(label)) {
        if (length(prot_seq) != length(label)) {
            stop("check the number of samples")
        }
    }
    
    check <- NULL
    for (i in seq_len(length(prot_seq))) {
        check[i] <- grepl("^[A-Z-]+$", prot_seq[i])
    }
    
    if (!all(check)) {
        result$removed_prot_seq <- which(!check)
        message("at least one of protein sequences may not be valid")
    }
    
    result$prot_seq <- prot_seq[check]
    result$label <- label[check]
    result
}

fit_GAN <- function(prot_seq,
                    label = NULL,
                    length_seq,
                    embedding_dim,
                    embedding_args = list(),
                    latent_dim = NULL,
                    intermediate_generator_layers,
                    intermediate_discriminator_layers,
                    prot_seq_val = NULL,
                    label_val = NULL,
                    epochs,
                    batch_size,
                    preprocessing = list(
                        x_train = NULL,
                        x_val = NULL,
                        y_train = NULL,
                        y_val = NULL,
                        lenc = NULL,
                        length_seq = NULL,
                        num_seq = NULL,
                        embedding_dim = NULL,
                        embedding_matrix = NULL,
                        removed_prot_seq = NULL,
                        removed_prot_seq_val = NULL,
                        latent_dim = NULL),
                    optimizer = "adam",
                    validation_split = 0) {
    result <- NULL
    result$preprocessing <- NULL
    
    ### pre-processing
    tryCatch({
        if (all(unlist(lapply(preprocessing, Negate(is.null))))) {
            result$preprocessing <- preprocessing
            x_train <- preprocessing$x_train
            if (is.null(x_train)) stop("check embedded sequence data for train in the preprocessed result")
            y_train <- preprocessing$y_train
            x_val <- preprocessing$x_val
            y_val <- preprocessing$y_val
            length_seq <- preprocessing$length_seq
            if (is.null(length_seq)) stop("check length of sequence in the preprocessed result")
            num_seq <- preprocessing$num_seq
            if (is.null(num_seq)) stop("check number of sequences for train in the preprocessed result")
            num_seq_val <- preprocessing$num_seq_val
            embedding_dim <- preprocessing$embedding_dim
            if (is.null(embedding_dim)) stop("check dimension of the dense embedding in the preprocessed result")
            embedding_matrix <- preprocessing$embedding_matrix
            if (is.null(embedding_matrix)) stop("check embedding matrix in the preprocessed result")
            if (is.null(latent_dim)) latent_dim <- preprocessing$latent_dim
        }
        
        if (any(unlist(lapply(preprocessing, is.null)))) {
            message("pre-processing...")
            
            # check
            checked_prot_seq <- prot_seq_check(prot_seq = prot_seq, label = label)
            prot_seq <- checked_prot_seq$prot_seq
            label <- checked_prot_seq$label
            result$preprocessing$removed_prot_seq <- checked_prot_seq$removed_prot_seq
            
            # sequence to vector
            prot2vec_train <- do.call(prot2vec,
                                    c(list(prot_seq = prot_seq, embedding_dim = embedding_dim),
                                        embedding_args))
            x_train <- prot2vec_train$prot_vec
            num_seq <- dim(x_train)[1]
            y_train <- label
            embedding_matrix <- prot2vec_train$embedding_matrix
            
            x_val <- NULL
            y_val <- NULL
            num_seq_val <- NULL
            if (!is.null(prot_seq_val)) {
                # check
                checked_prot_seq_val <- prot_seq_check(prot_seq = prot_seq_val, label = label_val)
                prot_seq_val <- checked_prot_seq_val$prot_seq
                label_val <- checked_prot_seq_val$label
                result$preprocessing$removed_prot_seq_val <- checked_prot_seq_val$removed_prot_seq
                
                # sequence to vector
                prot2vec_val <- prot2vec(prot_seq = prot_seq_val,
                                        embedding_matrix = embedding_matrix)
                x_val <- prot2vec_val$prot_vec
                y_val <- label_val
                num_seq_val <- dim(x_val)[1]
            }
            
            if (is.null(prot_seq_val) & validation_split) {
                x <- x_train
                idx <- sample(seq_len(nrow(x)))
                train_idx <- seq_len(nrow(x)) %in%
                    idx[seq_len(round(nrow(x) * (1 - validation_split)))]
                x_train <- x[train_idx,,]
                x_val <- x[!train_idx,,]
                if (!is.null(y_train) & is.null(y_val)) {
                    y <- y_train
                    y_train <- y[train_idx]
                    y_val <- y[!train_idx]
                }
                num_seq <- dim(x_train)[1]
                num_seq_val <- dim(x_val)[1]
            }
        }
    },
    error = function(e)
        message("In pre-processing : ", e)
    )
    
    
    ### building model
    tryCatch({
        # GAN
        g_model <- function(latent_dim, label) {
            generator_input <- layer_input(shape = c(latent_dim))
            if (!is.null(label)) {
                input_latent <- layer_input(shape = c(latent_dim))
                input_label <- layer_input(shape = c(1))
                embedding_label <- input_label %>%
                    layer_embedding(input_dim = length(unique(label)), output_dim = latent_dim) %>%
                    layer_flatten()
                generator_input <- layer_multiply(list(input_latent, embedding_label))
            }
            
            x <- generator_input
            for (i in seq_len(length(intermediate_generator_layers))) {
                x <- intermediate_generator_layers[[i]](x)
            }
            
            generator_output <- x %>%
                layer_dense(units = length_seq * embedding_dim) %>%
                layer_reshape(c(length_seq, embedding_dim))
            
            if (is.null(label)) {
                generator <- keras_model(generator_input, generator_output)
            } else {
                generator <- keras_model(list(input_latent, input_label), generator_output)
            }
        }
        
        d_model <- function(length_seq, embedding_dim, label, optimizer) {
            discriminator_input <- layer_input(shape = c(length_seq, embedding_dim))
            discriminator_output <- discriminator_input %>%
                layer_flatten(input_shape = c(length_seq, embedding_dim))
            
            x <- discriminator_output
            for (i in seq_len(length(intermediate_discriminator_layers))) {
                x <- intermediate_discriminator_layers[[i]](x)
            }
            
            if (is.null(label)) {
                discriminator_output <- x %>%
                    layer_dense(units = 1, activation = "sigmoid")
                discriminator <- keras_model(discriminator_input, discriminator_output)
            } else {
                fake <- discriminator_output %>% 
                    layer_dense(units = 1, activation = "sigmoid", name = "generation")
                aux <- discriminator_output %>%
                    layer_dense(units = length(unique(label)), activation = "softmax", name = "auxiliary")
                discriminator <- keras_model(discriminator_input, list(fake, aux))
            }
            
            if (is.null(label)) {
                discriminator %>% compile(
                    optimizer = optimizer,
                    loss = "binary_crossentropy"
                )        
            } else {
                discriminator %>% compile(
                    optimizer = optimizer,
                    loss = list("binary_crossentropy", "sparse_categorical_crossentropy")
                )
            }
        }
        
        combined_model <- function(g_model, d_model, label, optimizer) {
            freeze_weights(d_model)
            
            if (is.null(label)) {
                gan <- keras_model_sequential() %>% g_model %>% d_model
            } else {
                input_latent <- layer_input(shape = c(latent_dim))
                input_label <- layer_input(shape = c(1))
                results <- g_model(list(input_latent, input_label)) %>% d_model
                gan <- keras_model(list(input_latent, input_label), results)
            }
            
            gan_optimizer <- optimizer_adam()
            
            if (is.null(label)) {
                gan %>% compile(
                    optimizer = optimizer, 
                    loss = list("binary_crossentropy")
                )
            } else {
                gan %>% compile(
                    optimizer = optimizer,
                    loss = list("binary_crossentropy", "sparse_categorical_crossentropy")
                )
            }
        }
        
        generator <- g_model(latent_dim, y_train)
        discriminator <- d_model(length_seq, embedding_dim, y_train, optimizer)
        combined <- combined_model(generator, discriminator, y_train, optimizer)
    },
        error = function(e)
        message("In building model : ", e)
    )
    
    
    ### training
    tryCatch({
        message("training...")
        if (!is.null(y_train)) {
            if (any(unlist(lapply(preprocessing, is.null)))) {
                lenc <- CatEncoders::LabelEncoder.fit(y_train)
                y_train <- CatEncoders::transform(lenc, y_train) - 1
                result$preprocessing$lenc <- lenc
                if (!is.null(y_val)) {
                    y_val <- CatEncoders::transform(lenc, y_val) - 1
                }
            }
        }
        
        # epoch
        for (epoch in seq_len(epochs)) {
            epoch_gen_loss <- NULL
            epoch_disc_loss <- NULL
            
            num_batches <- trunc(num_seq/batch_size)
            possible_indexes <- seq_len(num_seq)
            
            cat(sprintf("Epoch %s/%s \n", epoch, epochs))
            progbar <- tensorflow::tf$keras$utils$Progbar(target = num_batches)
            
            # batch
            for (index in seq_len(num_batches)) {
                batch <- sample(possible_indexes, size = batch_size)
                possible_indexes <- possible_indexes[!possible_indexes %in% batch]
                
                
                # train discriminator
                noise <- matrix(stats::rnorm(batch_size * latent_dim), 
                                nrow = batch_size, ncol = latent_dim)
                x_train_batch <- x_train[batch,,,drop = FALSE]
                
                if (!is.null(y_train)) {
                    y_train_batch <- y_train[batch]
                    y_train_sample <- sample(unique(y_train), batch_size, replace = TRUE)
                    aux_y <- matrix(c(y_train_sample, y_train_batch), ncol = 1)
                    x_gen_batch <- predict(generator, list(noise, y_train_sample))
                } else {
                    x_gen_batch <- predict(generator, noise)
                }
                
                X <- array(0, dim = c(2 * batch_size, length_seq, embedding_dim))
                X[seq(batch_size),,] <- x_gen_batch
                X[seq(batch_size + 1, 2 * batch_size),,] <- x_train_batch
                
                fake_real <- c(rep(0, batch_size), rep(1, batch_size))
                
                if (!is.null(y_train)) {
                    disc_loss <- train_on_batch(
                        discriminator,
                        x = X,
                        y = list(fake_real, aux_y)
                    )
                } else {
                    disc_loss <- train_on_batch(
                        discriminator,
                        x = X,
                        y = fake_real
                    )
                }
                epoch_disc_loss <- rbind(epoch_disc_loss, unlist(disc_loss))
                
                
                # train generator
                noise <- matrix(stats::rnorm(2 * batch_size * latent_dim), 
                                nrow = 2 * batch_size, ncol = latent_dim)
                
                if (!is.null(y_train)) {
                    y_train_sample <- matrix(
                        sample(unique(y_train), size = 2 * batch_size, replace = TRUE),
                        ncol = 1)
                }
                
                misleading_real <- rep(1, 2 * batch_size)
                
                if (!is.null(y_train)) {
                    combined_loss <- train_on_batch(
                        combined,
                        list(noise, y_train_sample),
                        list(misleading_real, y_train_sample)
                    )
                } else {
                    combined_loss <- train_on_batch(
                        combined,
                        noise,
                        misleading_real)
                }
                epoch_gen_loss <- rbind(epoch_gen_loss, unlist(combined_loss))
                progbar$add(1)
            }
            
            discriminator_train_loss <- apply(epoch_disc_loss, 2, mean)
            generator_train_loss <- apply(epoch_gen_loss, 2, mean)
            
            if (!is.null(x_val)) {
                
                # evaluate discriminator
                noise <- matrix(stats::rnorm(num_seq_val * latent_dim), 
                                nrow = num_seq_val, ncol = latent_dim)
                
                if (!is.null(y_val)) {
                    y_val_sample <- sample(unique(y_val), size = num_seq_val, replace = TRUE)
                    aux_y <- matrix(c(y_val_sample, y_val), ncol = 1)
                    x_gen <- predict(generator, list(noise, y_val_sample))
                } else {
                    x_gen <- predict(generator, noise)
                }
                
                X <- array(0, dim = c(2 * num_seq_val, length_seq, embedding_dim))
                X[seq(num_seq_val),,] <- x_gen
                X[seq(num_seq_val + 1, 2 * num_seq_val),,] <- x_val
                
                fake_real <- c(rep(0, num_seq_val), rep(1, num_seq_val))
                
                if (!is.null(y_val)) {
                    discriminator_test_loss <- evaluate(
                        discriminator,
                        x = X,
                        y = list(fake_real, aux_y),
                        verbose = FALSE
                    ) %>% unlist()
                } else {
                    discriminator_test_loss <- evaluate(
                        discriminator,
                        x = X,
                        y = fake_real,
                        verbose = FALSE
                    ) %>% unlist()
                }
                
                
                # evaluate generator
                noise <- matrix(stats::rnorm(2 * num_seq_val * latent_dim), 
                                nrow = 2 * num_seq_val, ncol = latent_dim)
                
                if (!is.null(y_val)) {
                    y_val_sample <- matrix(
                        sample(unique(y_val), size = 2 * num_seq_val, replace = TRUE),
                        ncol = 1)
                }
                
                misleading_real <- rep(1, 2 * num_seq_val)
                
                if (!is.null(y_val)) {
                    generator_test_loss <- combined %>% evaluate(
                        list(noise, y_val_sample),
                        list(misleading_real, y_val_sample),
                        verbose = FALSE
                    )
                } else {
                    generator_test_loss <- combined %>% evaluate(
                        noise,
                        misleading_real,
                        verbose = FALSE
                    )
                }
            }
            
            
            # performance
            if (length(discriminator$metrics_names) == 1) {
                row_fmt <- "%s : loss %f \n"
            } else {
                row_fmt <- "%s : loss %f | generation_loss %f | auxiliary_loss %f \n"
            }
            
            cat(do.call(sprintf, c(list(row_fmt, "generator (train)"), generator_train_loss)))
            if (!is.null(x_val)) {
                cat(do.call(sprintf, c(list(row_fmt, "generator (test)"), generator_test_loss)))
            }
            cat(do.call(sprintf, c(list(row_fmt, "discriminator (train)"), discriminator_train_loss)))
            if (!is.null(x_val)) {
                cat(do.call(sprintf, c(list(row_fmt, "discriminator (test)"), discriminator_test_loss)))
            }
            cat("\n")
        }
    },
    error = function(e)
        message("In training : ", e)
    )
    
    result$model <- combined
    result$generator <- generator
    result$discriminator <- discriminator
    result$preprocessing$x_train <- x_train
    result$preprocessing$y_train <- y_train
    result$preprocessing$x_val <- x_val
    result$preprocessing$y_val <- y_val
    result$preprocessing$length_seq <- length_seq
    result$preprocessing$num_seq <- num_seq
    result$preprocessing$num_seq_val <- num_seq_val
    result$preprocessing$embedding_dim <- embedding_dim
    result$preprocessing$embedding_matrix <- embedding_matrix
    result$preprocessing$latent_dim <- latent_dim
    result
}





gen_GAN <- function(x, label = NULL, num_seq, remove_gap = TRUE) {
    result <- NULL
    latent_dim <- x$preprocessing$latent_dim
    if (is.null(latent_dim)) stop("check dimension of latent vector in the fitted result")
    
    ### generating
    tryCatch({
        message("generating...")
        if (is.null(x$preprocessing$y_train)) {
            gen_vec <- predict(x$generator,
                            matrix(stats::rnorm(num_seq * latent_dim),
                                    nrow = num_seq, ncol = latent_dim))
        } else {
            lenc <- x$preprocessing$lenc
            lable_enc <- CatEncoders::transform(lenc, label) - 1
            gen_vec <- predict(x$generator,
                            list(matrix(stats::rnorm(num_seq * latent_dim),
                                        nrow = num_seq, ncol = latent_dim),
                                        lable_enc))
        }
    },
        error = function(e)
        message("In generating : ", e)
    )
    
    
    ### post-processing
    tryCatch({
        message("post-processing...")
        # vector to sequence
        gen_seq <- vec2prot(gen_vec, x$preprocessing$embedding_matrix)
        
        if (remove_gap) {
            gen_seq <- gsub("-", "", gen_seq)
        } else {
            gen_seq <- gen_seq
        }
    },
        error = function(e)
        message("In post-processing : ", e)
    )
    
    result$gen_seq <- gen_seq
    result$label <- label
    result
}

# model parameters
length_seq <- 403
embedding_dim <- 8
latent_dim <- 4
epochs <- 20
batch_size <- 64



test_that("fit_GAN: miss training data", {
    expect_error(
        fit_GAN(length_seq = length_seq,
                embedding_dim = embedding_dim,
                latent_dim = latent_dim,
                intermediate_generator_layers = list(
                    layer_dense(units = 16),
                    layer_dense(units = 128)),
                intermediate_discriminator_layers = list(
                    layer_dense(units = 128, activation = "relu"),
                    layer_dense(units = 16, activation = "relu")),
                prot_seq_val = example_PTEN,
                epochs = epochs,
                batch_size = batch_size)
    )
})



test_that("fit_GAN: miss length of sequence", {
    expect_error(
        fit_GAN(prot_seq = example_PTEN,
                embedding_dim = embedding_dim,
                latent_dim = latent_dim,
                intermediate_generator_layers = list(
                    layer_dense(units = 16),
                    layer_dense(units = 128)),
                intermediate_discriminator_layers = list(
                    layer_dense(units = 128, activation = "relu"),
                    layer_dense(units = 16, activation = "relu")),
                prot_seq_val = example_PTEN,
                epochs = epochs,
                batch_size = batch_size)
    )
})



test_that("fit_GAN: miss embedding dimension", {
    expect_error(
        fit_GAN(prot_seq = example_PTEN,
                length_seq = length_seq,
                latent_dim = latent_dim,
                intermediate_generator_layers = list(
                    layer_dense(units = 16),
                    layer_dense(units = 128)),
                intermediate_discriminator_layers = list(
                    layer_dense(units = 128, activation = "relu"),
                    layer_dense(units = 16, activation = "relu")),
                prot_seq_val = example_PTEN,
                epochs = epochs,
                batch_size = batch_size)
    )
})



test_that("fit_GAN: miss latent dimension", {
    expect_error(
        fit_GAN(prot_seq = example_PTEN,
                length_seq = length_seq,
                embedding_dim = embedding_dim,
                intermediate_generator_layers = list(
                    layer_dense(units = 16),
                    layer_dense(units = 128)),
                intermediate_discriminator_layers = list(
                    layer_dense(units = 128, activation = "relu"),
                    layer_dense(units = 16, activation = "relu")),
                prot_seq_val = example_PTEN,
                epochs = epochs,
                batch_size = batch_size)
    )
})



test_that("fit_GAN: miss intermediate generator layers", {
    expect_error(
        fit_GAN(prot_seq = example_PTEN,
                length_seq = length_seq,
                embedding_dim = embedding_dim,
                latent_dim = latent_dim,
                intermediate_discriminator_layers = list(
                    layer_dense(units = 128, activation = "relu"),
                    layer_dense(units = 16, activation = "relu")),
                prot_seq_val = example_PTEN,
                epochs = epochs,
                batch_size = batch_size)
    )
})



test_that("fit_GAN: miss intermediate discriminator layers", {
    expect_error(
        fit_GAN(prot_seq = example_PTEN,
                length_seq = length_seq,
                embedding_dim = embedding_dim,
                latent_dim = latent_dim,
                intermediate_generator_layers = list(
                    layer_dense(units = 16),
                    layer_dense(units = 128)),
                prot_seq_val = example_PTEN,
                epochs = epochs,
                batch_size = batch_size)
    )
})



test_that("fit_GAN: miss epochs", {
    expect_error(
        fit_GAN(prot_seq = example_PTEN,
                length_seq = length_seq,
                embedding_dim = embedding_dim,
                latent_dim = latent_dim,
                intermediate_generator_layers = list(
                    layer_dense(units = 16),
                    layer_dense(units = 128)),
                intermediate_discriminator_layers = list(
                    layer_dense(units = 128, activation = "relu"),
                    layer_dense(units = 16, activation = "relu")),
                prot_seq_val = example_PTEN,
                batch_size = batch_size)
    )
})



test_that("fit_GAN: miss batch size", {
    expect_error(
        fit_GAN(prot_seq = example_PTEN,
                length_seq = length_seq,
                embedding_dim = embedding_dim,
                latent_dim = latent_dim,
                intermediate_generator_layers = list(
                    layer_dense(units = 16),
                    layer_dense(units = 128)),
                intermediate_discriminator_layers = list(
                    layer_dense(units = 128, activation = "relu"),
                    layer_dense(units = 16, activation = "relu")),
                prot_seq_val = example_PTEN,
                epochs = epochs)
    )
})



test_that("gen_GAN: miss result of fit_GAN", {
    expect_error(
        gen_GAN(num_seq = 100)
    )
})



test_that("fit_VAE: miss training data", {
    expect_error(
        fit_VAE(length_seq = length_seq,
                embedding_dim = embedding_dim,
                latent_dim = latent_dim,
                intermediate_encoder_layers = list(
                    layer_dense(units = 16),
                    layer_dense(units = 128)),
                intermediate_decoder_layers = list(
                    layer_dense(units = 128, activation = "relu"),
                    layer_dense(units = 16, activation = "relu")),
                prot_seq_val = example_PTEN,
                epochs = epochs,
                batch_size = batch_size)
    )
})



test_that("fit_VAE: miss length of sequence", {
    expect_error(
        fit_VAE(prot_seq = example_PTEN,
                embedding_dim = embedding_dim,
                latent_dim = latent_dim,
                intermediate_encoder_layers = list(
                    layer_dense(units = 16),
                    layer_dense(units = 128)),
                intermediate_decoder_layers = list(
                    layer_dense(units = 128, activation = "relu"),
                    layer_dense(units = 16, activation = "relu")),
                prot_seq_val = example_PTEN,
                epochs = epochs,
                batch_size = batch_size)
    )
})



test_that("fit_VAE: miss embedding dimension", {
    expect_error(
        fit_VAE(prot_seq = example_PTEN,
                length_seq = length_seq,
                latent_dim = latent_dim,
                intermediate_encoder_layers = list(
                    layer_dense(units = 16),
                    layer_dense(units = 128)),
                intermediate_decoder_layers = list(
                    layer_dense(units = 128, activation = "relu"),
                    layer_dense(units = 16, activation = "relu")),
                prot_seq_val = example_PTEN,
                epochs = epochs,
                batch_size = batch_size)
    )
})



test_that("fit_VAE: miss intermediate decoder layers", {
    expect_error(
        fit_VAE(prot_seq = example_PTEN,
                length_seq = length_seq,
                embedding_dim = embedding_dim,
                latent_dim = latent_dim,
                intermediate_encoder_layers = list(
                    layer_dense(units = 128, activation = "relu"),
                    layer_dense(units = 16, activation = "relu")),
                prot_seq_val = example_PTEN,
                epochs = epochs,
                batch_size = batch_size)
    )
})



test_that("fit_VAE: miss intermediate encoder layers", {
    expect_error(
        fit_VAE(prot_seq = example_PTEN,
                length_seq = length_seq,
                embedding_dim = embedding_dim,
                latent_dim = latent_dim,
                intermediate_decoder_layers = list(
                    layer_dense(units = 16),
                    layer_dense(units = 128)),
                prot_seq_val = example_PTEN,
                epochs = epochs,
                batch_size = batch_size)
    )
})



test_that("fit_VAE: miss epochs", {
    expect_error(
        fit_VAE(prot_seq = example_PTEN,
                length_seq = length_seq,
                embedding_dim = embedding_dim,
                latent_dim = latent_dim,
                intermediate_encoder_layers = list(
                    layer_dense(units = 16),
                    layer_dense(units = 128)),
                intermediate_decoder_layers = list(
                    layer_dense(units = 128, activation = "relu"),
                    layer_dense(units = 16, activation = "relu")),
                prot_seq_val = example_PTEN,
                batch_size = batch_size)
    )
})



test_that("fit_VAE: miss batch size", {
    expect_error(
      fit_VAE(prot_seq = example_PTEN,
              length_seq = length_seq,
              embedding_dim = embedding_dim,
              latent_dim = latent_dim,
              intermediate_encoder_layers = list(
                  layer_dense(units = 16),
                  layer_dense(units = 128)),
              intermediate_decoder_layers = list(
                  layer_dense(units = 128, activation = "relu"),
                  layer_dense(units = 16, activation = "relu")),
              prot_seq_val = example_PTEN,
              epochs = epochs)
    )
})



test_that("gen_VAE: miss result of fit_VAE", {
    expect_error(
        gen_VAE(num_seq = 100)
    )
})



length_seq <- 10
num_heads <- 2
ff_dim <- 16
num_transformer_blocks <- 2



test_that("fit_ART: miss training data", {
    expect_error(
        fit_ART(length_seq = length_seq,
                embedding_dim = embedding_dim,
                num_heads = num_heads,
                ff_dim = ff_dim,
                num_transformer_blocks = num_transformer_blocks,
                layers = list(layer_dropout(rate = 0.1),
                              layer_dense(units = 32, activation = "relu"),
                              layer_dropout(rate = 0.1)),
                epochs = epochs,
                batch_size = batch_size)
    )
})



test_that("fit_ART: invalid length of sequence", {
    expect_error(
        fit_ART(prot_seq = example_PTEN[1:5],
                length_seq = 500,
                embedding_dim = embedding_dim,
                num_heads = num_heads,
                ff_dim = ff_dim,
                num_transformer_blocks = num_transformer_blocks,
                layers = list(layer_dropout(rate = 0.1),
                              layer_dense(units = 32, activation = "relu"),
                              layer_dropout(rate = 0.1)),
                epochs = epochs,
                batch_size = batch_size)
    )
})



test_that("fit_ART: miss embedding dimension", {
    expect_error(
        fit_ART(prot_seq = example_PTEN[1:5],
                length_seq = length_seq,
                num_heads = num_heads,
                ff_dim = ff_dim,
                num_transformer_blocks = num_transformer_blocks,
                layers = list(layer_dropout(rate = 0.1),
                              layer_dense(units = 32, activation = "relu"),
                              layer_dropout(rate = 0.1)),
                epochs = epochs,
                batch_size = batch_size)
    )
})



test_that("fit_ART: miss num_heads", {
    expect_error(
        fit_ART(prot_seq = example_PTEN[1:5],
                length_seq = length_seq,
                embedding_dim = embedding_dim,
                ff_dim = ff_dim,
                num_transformer_blocks = num_transformer_blocks,
                layers = list(layer_dropout(rate = 0.1),
                              layer_dense(units = 32, activation = "relu"),
                              layer_dropout(rate = 0.1)),
                epochs = epochs,
                batch_size = batch_size)
    )
})



test_that("fit_ART: miss ff_dim", {
    expect_error(
        fit_ART(prot_seq = example_PTEN[1:5],
                length_seq = length_seq,
                num_heads = num_heads,
                embedding_dim = embedding_dim,
                num_transformer_blocks = num_transformer_blocks,
                layers = list(layer_dropout(rate = 0.1),
                              layer_dense(units = 32, activation = "relu"),
                              layer_dropout(rate = 0.1)),
                epochs = epochs,
                batch_size = batch_size)
    )
})



test_that("fit_ART: miss num_transformer_blocks", {
    expect_error(
        fit_ART(prot_seq = example_PTEN[1:5],
                length_seq = length_seq,
                num_heads = num_heads,
                embedding_dim = embedding_dim,
                ff_dim = ff_dim,
                layers = list(layer_dropout(rate = 0.1),
                              layer_dense(units = 32, activation = "relu"),
                              layer_dropout(rate = 0.1)),
                epochs = epochs,
                batch_size = batch_size)
    )
})



test_that("fit_ART: miss epochs", {
    expect_error(
        fit_ART(prot_seq = example_PTEN[1:5],
                length_seq = length_seq,
                num_heads = num_heads,
                embedding_dim = embedding_dim,
                ff_dim = ff_dim,
                num_transformer_blocks = num_transformer_blocks,
                layers = list(layer_dropout(rate = 0.1),
                              layer_dense(units = 32, activation = "relu"),
                              layer_dropout(rate = 0.1)),
                batch_size = batch_size)
    )
})



test_that("fit_ART: miss batch size", {
    expect_error(
        fit_ART(prot_seq = example_PTEN[1:5],
                length_seq = length_seq,
                num_heads = num_heads,
                embedding_dim = embedding_dim,
                ff_dim = ff_dim,
                num_transformer_blocks = num_transformer_blocks,
                layers = list(layer_dropout(rate = 0.1),
                              layer_dense(units = 32, activation = "relu"),
                              layer_dropout(rate = 0.1)),
                epochs = epochs)
    )
})



test_that("gen_ART: miss result of fit_ART", {
    expect_error(
        gen_ART(seed_prot = "SGFRKMAFPS", length_AA = 20, method = "greedy")
    )
})

fit_ART <- function(prot_seq,
                    length_seq,
                    embedding_dim,
                    num_heads,
                    ff_dim,
                    num_transformer_blocks,
                    layers = NULL,
                    prot_seq_val = NULL,
                    epochs,
                    batch_size,
                    preprocessing = list(
                        x_train = NULL,
                        x_val = NULL,
                        y_train = NULL,
                        y_val = NULL,
                        lenc = NULL,
                        length_seq = NULL,
                        num_AA = NULL,
                        embedding_dim = NULL,
                        removed_prot_seq = NULL,
                        removed_prot_seq_val = NULL),
                    use_generator = FALSE,
                    optimizer = "adam",
                    metrics = "accuracy",
                    validation_split = 0, ...) {
    result <- NULL
    result$preprocessing <- NULL
    x_train <- NULL
    x_val <- NULL
    y_train <- NULL
    y_val <- NULL
    
    ### pre-processing
    tryCatch({
        if (all(unlist(lapply(preprocessing, Negate(is.null))))) {
            result$preprocessing <- preprocessing
            x_train <- preprocessing$x_train
            if (is.null(x_train)) stop("check embedded sequence data for train in the preprocessed result")
            x_val <- preprocessing$x_val
            y_train <- preprocessing$y_train
            if (is.null(y_train)) stop("check labels for train in the preprocessed result")
            y_val <- preprocessing$y_val
            length_seq <- preprocessing$length_seq
            if (is.null(length_seq)) stop("check length of sequence in the preprocessed result")
            embedding_dim <- preprocessing$embedding_dim
            if (is.null(embedding_dim)) stop("check dimension of the dense embedding in the preprocessed result")
            lenc <- preprocessing$lenc
            if (is.null(lenc)) stop("check encoded labels in the preprocessed result")
            num_AA <- preprocessing$num_AA
            if (is.null(num_AA)) stop("check number of amino acids in the preprocessed result")
            if (embedding_dim %% num_heads != 0) {
                stop("the embedding dimension is a multiple of the number of attention heads")
            }
        } else {
            if (any(unlist(lapply(preprocessing, is.null)))) {
                message("pre-processing...")
                
                if (embedding_dim %% num_heads != 0) {
                    stop("the embedding dimension is a multiple of the number of attention heads")
                }
                checked_prot_seq <- prot_seq_check(prot_seq = prot_seq)
                prot_seq <- checked_prot_seq$prot_seq
                result$preprocessing$removed_prot_seq <- checked_prot_seq$removed_prot_seq
                
                temp <- NULL
                seq_encode_pad <- DeepPINCS::get_seq_encode_pad(prot_seq, max(nchar(prot_seq)))
                lenc <- seq_encode_pad$lenc
                num_AA <- seq_encode_pad$num_tokens
                for (i in seq_len(length(prot_seq))) {
                    temp <- rbind(temp, embed(rev(seq_encode_pad$sequences_encode_pad[i,][seq_encode_pad$sequences_encode_pad[i,] != 0]),
                                            length_seq+1))
                }
                x_train <- temp[, seq_len(length_seq)]
                y_train <- temp[, length_seq+1] - 1
                
                if (!is.null(prot_seq_val)) {
                    checked_prot_seq_val <- prot_seq_check(prot_seq = prot_seq_val)
                    prot_seq_val <- checked_prot_seq_val$prot_seq
                    result$preprocessing$removed_prot_seq_val <- checked_prot_seq_val$removed_prot_seq_val
                    
                    seq_val_encode_pad <- DeepPINCS::seq_preprocessing(AAseq = as.matrix(prot_seq_val),
                                                                        type = "sequence",
                                                                        length_seq = max(nchar(prot_seq_val)),
                                                                        lenc = lenc)
                    temp_val <- NULL
                    for (i in seq_len(length(prot_seq_val))) {
                        temp_val <- rbind(temp_val,
                                        embed(rev(seq_val_encode_pad$sequences_encode_pad[[1]][i,][seq_val_encode_pad$sequences_encode_pad[[1]][i,] != 0]),
                                            length_seq + 1))
                    }
                    x_val <- temp_val[, seq_len(length_seq)]
                    y_val <- temp_val[, length_seq+1] - 1
                }
                
                if (is.null(prot_seq_val) & validation_split) {
                    temp_x_train <- x_train
                    temp_y_train <- y_train
                    idx <- sample(seq_len(nrow(temp_x_train)))
                    train_idx <- seq_len(nrow(temp_x_train)) %in%
                        idx[seq_len(round(nrow(temp_x_train) * (1 - validation_split)))]
                    x_train <- temp_x_train[train_idx,]
                    x_val <- temp_x_train[!train_idx,]
                    y_train <- temp_y_train[train_idx]
                    y_val <- temp_y_train[!train_idx]
                }
            }
        }
    },
        error = function(e)
        message("In pre-processing : ", e)
    )
    
    
    ### building model
    tryCatch({
        # ART
        inputs <- layer_input(shape = length_seq)
        x <- inputs %>%
            layer_embedding_token_position(maxlen = length_seq,
                                            vocab_size = num_AA,
                                            embed_dim = embedding_dim) %>%
            layer_transformer_encoder(embed_dim = embedding_dim,
                                    num_heads = num_heads,
                                    ff_dim = ff_dim,
                                    num_transformer_blocks = num_transformer_blocks) %>%
            layer_global_average_pooling_1d()
        
        if (!is.null(layers)) {
            for (i in seq_len(length(layers))) {
                x <- layers[[i]](x)
            }
        }
        
        outputs <- layer_dense(units = num_AA, activation = "softmax")(x)
        model <- keras_model(inputs = inputs, outputs = outputs)
        model %>% compile(optimizer = optimizer,
                        loss = "sparse_categorical_crossentropy",
                        metrics = metrics)
    },
        error = function(e)
        message("In building model : ", e)
    )
    
    
    ### training
    tryCatch({
        message("training...")
        if (!use_generator) {
            # without generator
            validation_data <- NULL
            if (all(!is.null(x_val), !is.null(y_val))) {
                validation_data <- list(x_val, y_val)
            }
            
            model %>% keras::fit(x_train, y_train, epochs = epochs, batch_size = batch_size,
                                validation_data = validation_data,
                                validation_split = validation_split, ...)
        } else {
            # with generator
            validation_data <- NULL
            validation_steps <- NULL
            if (all(!is.null(x_val), !is.null(y_val))) {
                validation_data <- ttgsea::sampling_generator(
                    as.matrix(x_val), as.matrix(y_val),
                    batch_size = batch_size)
                validation_steps <- ceiling(nrow(x_val)/batch_size)
            }
            
            model %>% keras::fit(ttgsea::sampling_generator(as.matrix(x_train), 
                                                            as.matrix(y_train),
                                                            batch_size = batch_size),
                                    steps_per_epoch = nrow(x_train)/batch_size,
                                    epochs = epochs,
                                    validation_data = validation_data,
                                    validation_steps = validation_steps,...)
        }
    },
        error = function(e)
        message("In training : ", e)
    )
    
    result$preprocessing$x_train <- x_train
    result$preprocessing$y_train <- y_train
    result$preprocessing$x_val <- x_val
    result$preprocessing$y_val <- y_val
    result$preprocessing$length_seq <- length_seq
    result$preprocessing$embedding_dim <- embedding_dim
    result$preprocessing$lenc <- lenc
    result$preprocessing$num_AA <- num_AA
    result$model <- model
    result
}





gen_ART <- function(x, seed_prot, length_AA, method = NULL,
                    b = NULL, t = 1, k = NULL, p = NULL) {
    
    checked_seed_prot <- prot_seq_check(prot_seq = seed_prot)
    if (length(checked_seed_prot$prot_seq) == 0) stop("seed protein is not valid")
    model <- x$model
    if (is.null(model)) stop("check model in the fitted result")
    length_seq <- x$preprocessing$length_seq
    if (is.null(length_seq)) stop("check length of sequence in the fitted result")
    lenc <- x$preprocessing$lenc
    if (is.null(lenc)) stop("check encoded labels in the fitted result")
    y_train <- x$preprocessing$y_train
    if (is.null(y_train)) stop("check labels for train in the fitted result")
    labels <- sort(unique(y_train))
    
    
    ### generating
    tryCatch({
        message("generating...")
        temp_prot <- prot <- checked_seed_prot$prot_seq
        X <- DeepPINCS::get_seq_encode_pad(temp_prot, length_seq = length_seq, lenc = lenc)$sequences_encode_pad
        for (i in seq_len(length_AA)) {
            pred <- predict(model, rbind(X))
            if (method == "greedy") {
                temp_X <- apply(pred, 1, which.max)
            } else if (method == "beam") {
                temp_X <- apply(pred, 1,
                                function(x) {
                                    candidate <- labels[order(x, decreasing = TRUE)[seq_len(b)]] + 1
                                    candidate_prob <- x[order(x, decreasing = TRUE)[seq_len(b)]]
                                    candidate_next_prob <- 0
                                    for (j in seq_len(b)) {
                                        candidate_next_prob[j] <- max(predict(model, rbind(c(X[-1], candidate[j]))))
                                    }
                                    candidate[which.max(candidate_prob * candidate_next_prob)]
                                })
            } else if (method == "temperature") {
                temp_X <- apply(pred, 1,
                                function(x) {
                                    pred_temp <- log(x)/t
                                    pred_temp <- exp(pred_temp)/sum(exp(pred_temp))
                                    sample(length(x), size = 1, prob = pred_temp)
                                })
                
            } else if (method == "top_k") {
                temp_X <- apply(pred, 1,
                                function(x) sample(order(x, decreasing = TRUE)[seq_len(k)], size = 1,
                                                prob = x[order(x, decreasing = TRUE)[seq_len(k)]]/
                                                    sum(x[order(x, decreasing = TRUE)[seq_len(k)]])))
            } else if (method == "top_p") {
                temp_X <- apply(pred[1,,drop=FALSE], 1,
                                function(x) {
                                    k <- min(which((cumsum(x[order(x, decreasing = TRUE)]) > p) == TRUE))
                                    if (length(order(x, decreasing = TRUE)[seq_len(k)]) == 1) {
                                        order(x, decreasing = TRUE)[seq_len(k)]
                                    } else {
                                        sample(order(x, decreasing = TRUE)[seq_len(k)], size = 1,
                                            prob = x[order(x, decreasing = TRUE)[seq_len(k)]]/
                                                sum(x[order(x, decreasing = TRUE)[seq_len(k)]]))
                                    }
                                })
            } else {
                stop("method is not available")
            }
            
            temp_X <- labels[temp_X] + 1
            gen_AA <- CatEncoders::inverse.transform(lenc, temp_X)
            prot <- paste(prot, gen_AA, sep = "")
            temp_prot <- substr(prot, nchar(gen_AA)-length_seq+1, nchar(gen_AA))
            X <- c(X[-1], temp_X)
        }
        prot
    },
        error = function(e)
        message("In generating : ", e)
    )
}

prot2vec <- function(prot_seq, embedding_dim, embedding_matrix = NULL, ...) {
    result <- NULL
    num_seq <- length(prot_seq)
    AA_seq <- strsplit(prot_seq, "")
    
    if(is.null(embedding_matrix)) {
        model <- word2vec::word2vec(x = gsub("", " ", prot_seq), dim = embedding_dim, min_count = 1, split = " ", ...)
        embedding_matrix <- as.matrix(model)
    } else {
        embedding_matrix <- embedding_matrix
    }
    
    length_seq <- nchar(prot_seq[1])
    prot_vec <- array(0, dim = c(num_seq, length_seq, ncol(embedding_matrix)))
    
    for (i in seq_len(num_seq)) {
        prot_vec[i,,] <- embedding_matrix[AA_seq[[i]],]
    }
    
    result$prot_vec <- prot_vec
    result$embedding_matrix <- embedding_matrix
    result
}





vec2prot <- function(prot_vec, embedding_matrix) {
    prot_seq <- NULL
    num_seq <- dim(prot_vec)[1]
    length_seq <- dim(prot_vec)[2]
    for (i in seq_len(num_seq)) {
        temp_prot_seq <- 0
        for (j in seq_len(length_seq)) {
            similarity <- word2vec::word2vec_similarity(prot_vec[i,j,], embedding_matrix)
            sorted_similarity <- gsub("[^A-Z|-]", "", colnames(similarity)[order(similarity, decreasing = TRUE)])
            temp_prot_seq[j] <- sorted_similarity[sorted_similarity != ""][1]
        }
        prot_seq  <- c(prot_seq, paste(temp_prot_seq, collapse = ""))
    }
    prot_seq
}

layer_embedding_token_position <- function(x, maxlen, vocab_size, embed_dim) {
    layer_token_position_embedding <- Layer(
        classname = "TokenPositionEmbedding", 
        initialize = function(maxlen, vocab_size, embed_dim, ...) {
            super()$`__init__`(...)
            self$token_emb <- layer_embedding(
                input_dim = vocab_size + 1,
                output_dim = embed_dim,
                mask_zero = TRUE)
            self$pos_emb <- layer_embedding(input_dim = maxlen, output_dim = embed_dim)
        },
        call = function(inputs, ...) {
            maxlen <- tensorflow::tf$shape(inputs)[length(tensorflow::tf$shape(inputs))]
            positions <- tensorflow::tf$range(start = 0, limit = maxlen, delta = 1)
            positions <- self$pos_emb(positions)
            x <- self$token_emb(inputs)
            x + positions
        },
        get_config = function() {
            list(
                name = self$name
            )
        }
    )
    
    if (!is.null(globalenv()$maxlen)) old_maxlen <- maxlen
    if (!is.null(globalenv()$vocab_size)) old_vocab_size <- vocab_size
    if (!is.null(globalenv()$embed_dim)) old_embed_dim <- embed_dim
    
    assign("maxlen", maxlen, envir = globalenv())
    assign("vocab_size", vocab_size, envir = globalenv())
    assign("embed_dim", embed_dim, envir = globalenv())
    
    x <- layer_token_position_embedding(x,
                                        maxlen = maxlen,
                                        vocab_size = vocab_size,
                                        embed_dim = embed_dim)
    if (!exists("old_maxlen")) rm(maxlen, envir = globalenv())
    if (!exists("old_vocab_size")) rm(vocab_size, envir = globalenv())
    if (!exists("old_embed_dim")) rm(embed_dim, envir = globalenv())
    x
}





layer_transformer_encoder <- function(x, embed_dim, num_heads, ff_dim, num_transformer_blocks) {
    layer_MHA <- Layer(
        classname = "MultiHeadAttention", 
        initialize = function(embed_dim, num_heads, ...) {
            super()$`__init__`(...)
            self$embed_dim <- embed_dim
            self$num_heads <- num_heads
            
            stopifnot(embed_dim %% self$num_heads == 0)
            
            self$depth <- embed_dim %/% num_heads
            self$query_dense <- layer_dense(units = embed_dim)
            self$key_dense <- layer_dense(units = embed_dim)
            self$value_dense <- layer_dense(units = embed_dim)
            self$dense <- layer_dense(units = embed_dim)
        },
        scaled_dot_product_attention = function(query, key, value) {
            matmul_qk <- tensorflow::tf$matmul(query, key, transpose_b = TRUE)
            dk <- tensorflow::tf$cast(tensorflow::tf$shape(key)[length(tensorflow::tf$shape(key))],
                                    tensorflow::tf$float32)
            scaled_attention_logits <- matmul_qk / tensorflow::tf$math$sqrt(dk)
            attention_weights <- tensorflow::tf$nn$softmax(scaled_attention_logits, axis = -1)
            output <- tensorflow::tf$matmul(attention_weights, value)
            output
        },
        split_heads = function(x, batch_size) {
            x <- tensorflow::tf$reshape(x, list(batch_size, -1L,
                                                as.integer(self$num_heads),
                                                as.integer(self$depth)))
            tensorflow::tf$transpose(x, perm=c(0L, 2L, 1L, 3L))
        },
        call = function(inputs, ...) {
            batch_size <- tensorflow::tf$shape(inputs)[1]
            
            query <- self$query_dense(inputs)
            key <- self$key_dense(inputs)
            value <- self$value_dense(inputs)
            
            query <- self$split_heads(query, batch_size)  
            key <- self$split_heads(key, batch_size)
            value <- self$split_heads(value, batch_size)
            
            scaled_attention <- self$scaled_dot_product_attention(query, key, value)
            scaled_attention <- tensorflow::tf$transpose(scaled_attention, perm=c(0L, 2L, 1L, 3L))  
            
            concat_attention <- tensorflow::tf$reshape(scaled_attention,
                                                    list(batch_size, -1L,
                                                        as.integer(self$embed_dim)))
            outputs <- self$dense(concat_attention)
            outputs
        },
        get_config = function() {
            list(
                name = self$name
            )
        }
    )
    
    layer_transformer_block <- Layer(
        classname = "TransformerBlock", 
        initialize = function(embed_dim, num_heads, ff_dim, rate=0.1, ...) {
            super()$`__init__`(...)
            self$att <- layer_MHA(embed_dim=embed_dim, num_heads=num_heads)
            self$ffn <- keras_model_sequential() %>%
                layer_dense(units = ff_dim, activation = "relu") %>%
                layer_dense(units = embed_dim)
            self$layernorm1 <- layer_layer_normalization(epsilon = 1e-6)
            self$layernorm2 <- layer_layer_normalization(epsilon = 1e-6)
            self$dropout1 <- layer_dropout(rate = rate)
            self$dropout2 <- layer_dropout(rate = rate)
        },
        call = function(inputs, ...) {
            attn_output <- self$att(inputs)
            attn_output <- self$dropout1(attn_output)
            out1 <- self$layernorm1(inputs + attn_output)
            ffn_output <- self$ffn(out1)
            ffn_output <- self$dropout2(ffn_output)
            self$layernorm2(out1 + ffn_output)
        },
        get_config = function() {
            list(
                name = self$name
            )
        }
    )
    
    if (!is.null(globalenv()$embed_dim)) old_embed_dim <- embed_dim
    if (!is.null(globalenv()$num_heads)) old_num_heads <- num_heads
    if (!is.null(globalenv()$ff_dim)) old_ff_dim <- ff_dim
    
    assign("embed_dim", embed_dim, envir = globalenv())
    assign("num_heads", num_heads, envir = globalenv())
    assign("ff_dim", ff_dim, envir = globalenv())
    
    for (i in seq_len(num_transformer_blocks)) {
        x <- layer_transformer_block(x,
                                    embed_dim = embed_dim,
                                    num_heads = num_heads,
                                    ff_dim = ff_dim)
    }
    if (!exists("old_embed_dim")) rm(embed_dim, envir = globalenv())
    if (!exists("old_num_heads")) rm(num_heads, envir = globalenv())
    if (!exists("old_ff_dim")) rm(ff_dim, envir = globalenv())
    x
}

fit_VAE <- function(prot_seq,
                    label = NULL,
                    length_seq,
                    embedding_dim,
                    embedding_args = list(),
                    latent_dim = 2,
                    intermediate_encoder_layers,
                    intermediate_decoder_layers,
                    prot_seq_val = NULL,
                    label_val = NULL,
                    regularization = 1,
                    epochs,
                    batch_size,
                    preprocessing = list(
                        x_train = NULL,
                        x_val = NULL,
                        y_train = NULL,
                        y_val = NULL,
                        lenc = NULL,
                        length_seq = NULL,
                        embedding_dim = NULL,
                        embedding_matrix = NULL,
                        removed_prot_seq = NULL,
                        removed_prot_seq_val = NULL),
                    use_generator = FALSE,
                    optimizer = "adam",
                    validation_split = 0, ...) {
    if (tensorflow::tf$executing_eagerly())
        tensorflow::tf$compat$v1$disable_eager_execution()
    
    result <- NULL
    result$preprocessing <- NULL
    if (regularization < 0) {
        stop("regularization parameter should be nonnegative")
    }
    
    
    ### pre-processing
    tryCatch({
        if (all(unlist(lapply(preprocessing, Negate(is.null))))) {
            result$preprocessing <- preprocessing
            x_train <- preprocessing$x_train
            if (is.null(x_train)) stop("check embedded sequence data for train in the preprocessed result")
            y_train <- preprocessing$y_train
            x_val <- preprocessing$x_val
            y_val <- preprocessing$y_val
            length_seq <- preprocessing$length_seq
            if (is.null(length_seq)) stop("check length of sequence in the preprocessed result")
            embedding_dim <- preprocessing$embedding_dim
            if (is.null(embedding_dim)) stop("check dimension of the dense embedding in the preprocessed result")
            embedding_matrix <- preprocessing$embedding_matrix
            if (is.null(embedding_matrix)) stop("check embedding matrix in the preprocessed result")
        }
        
        if (any(unlist(lapply(preprocessing, is.null)))) {
            message("pre-processing...")
            
            # check
            checked_prot_seq <- prot_seq_check(prot_seq = prot_seq, label = label)
            prot_seq <- checked_prot_seq$prot_seq
            label <- checked_prot_seq$label
            result$preprocessing$removed_prot_seq <- checked_prot_seq$removed_prot_seq
            
            # sequence to vector
            prot2vec_train <- do.call(prot2vec,
                                    c(list(prot_seq = prot_seq, embedding_dim = embedding_dim),
                                        embedding_args))
            x_train <- prot2vec_train$prot_vec
            num_seq <- length(prot_seq)
            x_train <- reticulate::array_reshape(x_train, c(num_seq, length_seq * embedding_dim))
            y_train <- label
            embedding_matrix <- prot2vec_train$embedding_matrix
            
            x_val <- NULL
            y_val <- NULL
            if (!is.null(prot_seq_val)) {
                # check
                checked_prot_seq_val <- prot_seq_check(prot_seq = prot_seq_val, label = label_val)
                prot_seq_val <- checked_prot_seq_val$prot_seq
                label_val <- checked_prot_seq_val$label
                result$preprocessing$removed_prot_seq_val <- checked_prot_seq_val$removed_prot_seq
                # sequence to vector
                prot2vec_val <- prot2vec(prot_seq = prot_seq_val,
                                        embedding_matrix = embedding_matrix)
                x_val <- prot2vec_val$prot_vec
                x_val <- reticulate::array_reshape(x_val, c(num_seq, length_seq * embedding_dim))
                y_val <- label_val
            }
            
            if (is.null(prot_seq_val) & validation_split) {
                x <- x_train
                idx <- sample(seq_len(nrow(x)))
                train_idx <- seq_len(nrow(x)) %in%
                    idx[seq_len(round(nrow(x) * (1 - validation_split)))]
                x_train <- x[train_idx,]
                x_val <- x[!train_idx,]
                if (!is.null(y_train) & is.null(y_val)) {
                    y <- y_train
                    y_train <- y[train_idx]
                    y_val <- y[!train_idx]
                }
            }
        }
    },
        error = function(e)
        message("In pre-processing : ", e)
    )
    
    
    ### building model
    tryCatch({
        # VAE
        x <- layer_input(shape = c(length_seq * embedding_dim))
        if (is.null(y_train)) {
            encoded <- x
        } else {
            condition <- layer_input(shape = c(1), name = "condition")
            encoded <- layer_concatenate(c(x, condition))
        }
        
        for (i in seq_len(length(intermediate_encoder_layers))) {
            encoded <- intermediate_encoder_layers[[i]](encoded)
        }
        
        assign("z_mean", layer_dense(encoded, latent_dim, name = "z_mean"),
            envir = globalenv())
        assign("z_log_stddev", layer_dense(encoded, latent_dim, name = "z_log_stddev"),
            envir = globalenv())
        
        z <- layer_lambda(list(z_mean, z_log_stddev),
                        function(arg) {
                            z_mean <- arg[[1]]
                            z_log_stddev <- arg[[2]]
                            epsilon <- k_random_normal(shape = c(k_shape(z_mean)[[1]]),
                                                        mean = 0.0, stddev = 1)
                            z_mean + k_exp(z_log_stddev)*epsilon
                        }, name = "latent")
        
        if (is.null(y_train)) {
            decoded <- z
        } else {
            decoded <- layer_concatenate(c(z, condition))
        }
        decoder_layers <- c(intermediate_decoder_layers,
                            layer_dense(units = length_seq * embedding_dim))
        for (i in seq_len(length(decoder_layers))) {
            decoded <- decoder_layers[[i]](decoded)
        }
        
        if (is.null(y_train)) {
            model <- keras_model(inputs = x, outputs = decoded)
        } else {
            model <- keras_model(inputs = c(x, condition), outputs = decoded)
        }
        
        # encoder
        if (is.null(y_train)) {
            encoder <- keras_model(inputs = x, outputs = z_mean)
        } else {
            encoder <- keras_model(inputs = c(x, condition), outputs = z_mean)
        }
        
        # decoder
        if (is.null(y_train)) {
            decoder_input <- layer_input(shape = latent_dim)
        } else {
            decoder_input <- layer_input(shape = latent_dim + 1)
        }
        decoder_output <- decoder_input
        for (i in seq_len(length(decoder_layers))) {
            decoder_output <- decoder_layers[[i]](decoder_output)
        }
        decoder <- keras_model(decoder_input, decoder_output)
        
        # loss
        vae_loss <- function(x_true, x_pred) {
            xent_loss <- (length_seq * embedding_dim / 1.0) * loss_mean_squared_error(x_true, x_pred)
            kl_loss <- -0.5 * k_mean(1 + z_log_stddev - k_square(z_mean) - k_exp(z_log_stddev), axis = -1L)
            xent_loss + regularization * kl_loss
        }
        
        model %>% keras::compile(optimizer = optimizer, loss = vae_loss)
    },
        error = function(e)
        message("In building model : ", e)
    )
    
    
    ### training
    tryCatch({
        message("training...")
        if (is.null(y_train)) {
            # vanilla vae
            if (!use_generator) {
                # without generator
                validation_data <- NULL
                if (!is.null(x_val)) {
                    validation_data <- list(x_val, x_val)
                }
                
                model %>% keras::fit(
                    x_train, x_train,
                    epochs = epochs, 
                    batch_size = batch_size, 
                    validation_data = validation_data,
                    validation_split = validation_split, ...)
            } else {
                # with generator
                validation_data <- NULL
                validation_steps <- NULL
                if (!is.null(x_val)) {
                    validation_data <- DeepPINCS::multiple_sampling_generator(
                        list(x_val), x_val,
                        batch_size = batch_size)
                    validation_steps <- ceiling(nrow(x_val)/batch_size)
                }
                
                model %>% keras::fit(
                    DeepPINCS::multiple_sampling_generator(
                        list(x_train), x_train,
                        batch_size = batch_size),
                    steps_per_epoch = ceiling(nrow(x_train)/batch_size),
                    epochs = epochs,
                    validation_data = validation_data,
                    validation_steps = validation_steps, ...)
            }
        } else {
            # conditional vae
            if (any(unlist(lapply(preprocessing, is.null)))) {
                lenc <- CatEncoders::LabelEncoder.fit(y_train)
                y_train <- CatEncoders::transform(lenc, y_train)
                result$preprocessing$lenc <- lenc
                
                if (!is.null(y_val)) {
                    y_val <- CatEncoders::transform(lenc, y_val)
                }
            }
            
            if (!use_generator) {
                # without generator
                validation_data <- NULL
                if (!is.null(x_val)) {
                    validation_data <- list(list(x_val, y_val), x_val)
                }
                
                model %>% keras::fit(
                    list(x_train, y_train), x_train,
                    shuffle = TRUE, 
                    epochs = epochs, 
                    batch_size = batch_size, 
                    validation_data = validation_data,
                    validation_split = validation_split, ...)
            } else {
                # with generator
                validation_data <- NULL
                validation_steps <- NULL
                if (!is.null(x_val)) {
                    validation_data <- DeepPINCS::multiple_sampling_generator(
                        list(x_val, cbind(y_val)), cbind(x_val),
                        batch_size = batch_size)
                    validation_steps <- ceiling(nrow(x_val)/batch_size)
                }
                
                model %>% keras::fit(
                    DeepPINCS::multiple_sampling_generator(
                        list(x_train, cbind(y_train)), cbind(x_train),
                        batch_size = batch_size),
                    steps_per_epoch = ceiling(nrow(x_train)/batch_size),
                    epochs = epochs,
                    validation_data = validation_data,
                    validation_steps = validation_steps, ...)
            }
        }
    },
        error = function(e)
        message("In training : ", e)
    )
    
    rm(z_mean, envir = globalenv())
    rm(z_log_stddev, envir = globalenv())
    
    result$model <- model
    result$encoder <- encoder
    result$decoder <- decoder
    result$preprocessing$x_train <- x_train
    result$preprocessing$y_train <- y_train
    result$preprocessing$x_val <- x_val
    result$preprocessing$y_val <- y_val
    result$preprocessing$length_seq <- length_seq
    result$preprocessing$embedding_dim <- embedding_dim
    result$preprocessing$embedding_matrix <- embedding_matrix
    result
}





gen_VAE <- function(x, label = NULL, num_seq, remove_gap = TRUE,
                    batch_size, use_generator = FALSE) {
    result <- NULL
    encoder <- x$encoder
    if (is.null(encoder)) stop("check encoder model in the fitted result")
    decoder <- x$decoder
    if (is.null(decoder)) stop("check decoder model in the fitted result")
    x_train <- x$preprocessing$x_train
    if (is.null(x_train)) stop("check embedded sequence data for train in the fitted result")
    y_train <- x$preprocessing$y_train
    length_seq <- x$preprocessing$length_seq
    if (is.null(length_seq)) stop("check length of sequence in the fitted result")
    embedding_dim <- x$preprocessing$embedding_dim
    if (is.null(embedding_dim)) stop("check dimension of the dense embedding in the fitted result")
    
    x_train_encoded <- predict(encoder, list(x_train, y_train))
    rownames(x_train_encoded) <- rownames(x_train)
    colnames(x_train_encoded) <- paste("latent", seq_len(ncol(x_train_encoded)))
    
    
    ### generating
    tryCatch({
        message("generating...")
        if (is.null(y_train)) {
            # vanilla vae
            label <- NULL
            model_BIC <- mclust::mclustBIC(x_train_encoded, verbose = FALSE)
            mod <- mclust::mclustModel(x_train_encoded, model_BIC)
            z_sample <- mclust::sim(modelName = mod$modelName, 
                                    parameters = mod$parameters, 
                                    n = num_seq)
            
            if (!use_generator) {
                x_gen <- predict(decoder, z_sample[,-1])
            } else {
                x_gen <- predict(decoder,
                                DeepPINCS::multiple_sampling_generator(
                                    list(z_sample[,-1]),
                                    batch_size = batch_size,
                                    shuffle = FALSE),
                                steps = ceiling(nrow(z_sample)/batch_size))
            }
            
        } else {
            # conditional vae
            lenc <- x$preprocessing$lenc
            label_enc <- CatEncoders::transform(lenc, label)
            all_labels <- names(table(label_enc))
            z_sample <- matrix(0, num_seq, ncol(x_train_encoded))
            for (i in seq_len(length(all_labels))) {
                model_BIC <- mclust::mclustBIC(x_train_encoded[y_train == all_labels[i],],
                                            verbose = FALSE)
                mod <- mclust::mclustModel(x_train_encoded[y_train == all_labels[i],], model_BIC)
                temp_z <- mclust::sim(modelName = mod$modelName, 
                                    parameters = mod$parameters, 
                                    n = table(label_enc)[i])
                z_sample[label_enc == all_labels[i],] <- temp_z[,-1]
            }
            
            if (!use_generator) {
                gen_vec <- predict(decoder, cbind(z_sample, label_enc))
            } else {
                gen_vec <- predict(decoder,
                                    DeepPINCS::multiple_sampling_generator(
                                        list(cbind(z_sample, label_enc)),
                                        batch_size = batch_size,
                                        shuffle = FALSE),
                                    steps = ceiling(nrow(z_sample)/batch_size))
            }
        }
    },
        error = function(e)
        message("In generating : ", e)
    )
    
    
    ### post-processing
    tryCatch({
        message("post-processing...")
        # vector to sequence
        gen_vec <- reticulate::array_reshape(as.matrix(gen_vec),
                                            c(num_seq, length_seq, embedding_dim))
        gen_seq <- vec2prot(gen_vec, x$preprocessing$embedding_matrix)
        
        if(remove_gap) {
            gen_seq <- gsub("-", "", gen_seq)
        } else {
            gen_seq <- gen_seq
        }
    },
        error = function(e)
        message("In post-processing : ", e)
    )
    
    result$gen_seq <- gen_seq
    result$label <- label
    result$latent_vector <- x_train_encoded
    result
}

\name{prot_vec}
\alias{prot2vec}
\alias{vec2prot}

\title{
Converting from protein sequences to vectors or vice versa.
}

\description{
By using the word2vec model, amino acids are mapped to vectors of real numbers. Conceptually, it involves a mathematical embedding from a space with many dimensions per amino acid to a continuous vector space with a much lower dimension.
}

\usage{
prot2vec(prot_seq, embedding_dim, embedding_matrix = NULL, ...)
vec2prot(prot_vec, embedding_matrix)
}

\arguments{
    \item{prot_seq}{
protein sequences
}
    \item{prot_vec}{
protein embedding vectors
}
    \item{embedding_dim}{
dimension of embedding vectors
}
    \item{embedding_matrix}{
embedding matrix (default: NULL)
}
    \item{...}{
arguments for "word2vec::word2vec" but for dim, min_count and split 
}
}


\value{
    \item{prot_seq}{
protein sequences
}
    \item{prot_vec}{
protein embedding vectors
}
    \item{embedding_matrix}{
embedding matrix
}
}


\author{
Dongmin Jung
}


\seealso{
word2vec::word2vec, word2vec::word2vec_similarity
}


\references{
Mikolov, T., Sutskever, I., Chen, K., Corrado, G., & Dean, J. (2013). Distributed Representations of Words and Phrases and their Compositionality. arXiv:1310.4546.

Chang, M. (2020). Artificial intelligence for drug development, precision medicine, and healthcare.
}


\examples{
data("example_PTEN")
prot_seq <- example_PTEN[1:10]
prot2vec_result <- prot2vec(prot_seq = prot_seq, embedding_dim = 8)
vec2prot_result <- vec2prot(prot_vec = prot2vec_result$prot_vec,
                            embedding_matrix = prot2vec_result$embedding_matrix)
}

\name{transformer}
\alias{layer_embedding_token_position}
\alias{layer_transformer_encoder}

\title{
Transformer model
}

\description{
The Transformer architecture is a nonrecurrent architecture with a series of attention-based blocks. Each block is composed of a multi-head attention layer and a position-wise feedforward layer with an add and normalize layer in between. These layers process input sequences simultaneously, in parallel, independently of sequential order.
}

\usage{
layer_embedding_token_position(x, maxlen, vocab_size, embed_dim)
layer_transformer_encoder(x, embed_dim, num_heads, ff_dim, num_transformer_blocks)
}

\arguments{
    \item{x}{
layer object
}
    \item{maxlen}{
maximum of sequence size
}
    \item{vocab_size}{
vacabulary size
}
    \item{embed_dim}{
embedding size for each token
}
    \item{num_heads}{
number of attention heads
}
    \item{ff_dim}{
hidden layer size in feedforward network inside transformer
}
    \item{num_transformer_blocks}{
number of transformer blocks
}
}


\value{
layer object
}


\author{
Dongmin Jung
}


\references{
Lappin, S. (2021). Deep learning and linguistic representation. CRC Press.

Liu, Z., Lin, Y., & Sun, M. (2020). Representation learning for natural language processing. Springer.
}


\examples{
if (keras::is_keras_available() & reticulate::py_available()) {
    num_AA <- 20
    length_seq <- 10
    embedding_dim <- 16
    num_heads <- 2
    ff_dim <- 16
    num_transformer_blocks <- 2
    
    inputs <- layer_input(shape = length_seq)
    x <- inputs \%>\%
        layer_embedding_token_position(maxlen = length_seq,
                                    vocab_size = num_AA,
                                    embed_dim = embedding_dim) \%>\%
        layer_transformer_encoder(embed_dim = embedding_dim,
                                num_heads = num_heads,
                                ff_dim = ff_dim,
                                num_transformer_blocks = num_transformer_blocks) \%>\%
        layer_global_average_pooling_1d()
}
}

\name{example_PTEN}
\alias{example_PTEN}

\title{
Example Data for Protein Sequences
}

\description{
The data consist of selected amino acid sequences of the PTEN. There are 912 aligned sequences of length 403.
}

\usage{
example_PTEN
}

\value{
aligned amino acid sequences
}

\source{
Frazer, J., Notin, P., Dias, M., Gomez, A., Brock, K., Gal, Y., & Marks, D. (2020). Large-scale clinical interpretation of genetic variants using evolutionary data and deep learning. bioRxiv.

}

\author{
Dongmin Jung
}

\name{ART}
\alias{fit_ART}
\alias{gen_ART}

\title{
Autoregressive language model with Transformer
}

\description{
The autoregressive generative model predicts the next amino acid in a protein given the amino acid sequence up to that point. The autoregressive model generates proteins one amino acid at a time. For one step of generation, it takes a context sequence of amino acids as input and outputs a probability distribution over amino acids. We sample from that distribution and then update the context sequence with the sampled amino acid. The Transformer is used as an encoder model. The AR with the Transformer model can be trained by the function "fit_ART", and then the function "gen_ART" generates protein sequences.
}

\usage{
fit_ART(prot_seq,
        length_seq,
        embedding_dim,
        num_heads,
        ff_dim,
        num_transformer_blocks,
        layers = NULL,
        prot_seq_val = NULL,
        epochs,
        batch_size,
        preprocessing = list(
            x_train = NULL,
            x_val = NULL,
            y_train = NULL,
            y_val = NULL,
            lenc = NULL,
            length_seq = NULL,
            num_AA = NULL,
            embedding_dim = NULL,
            removed_prot_seq = NULL,
            removed_prot_seq_val = NULL),
        use_generator = FALSE,
        optimizer = "adam",
        metrics = "accuracy",
        validation_split = 0, ...) 

gen_ART(x,
        seed_prot,
        length_AA,
        method = NULL,
        b = NULL,
        t = 1,
        k = NULL,
        p = NULL)
}

\arguments{
    \item{prot_seq}{
amino acid sequence
}
    \item{length_seq}{
length of sequence used as input
}
    \item{embedding_dim}{
dimension of the dense embedding
}
    \item{num_heads}{
number of attention heads
}
    \item{ff_dim}{
hidden layer size in feedforward network inside transformer
}
    \item{num_transformer_blocks}{
number of transformer blocks
}
    \item{layers}{
list of layers between the transformer encoder and the output layer (default: NULL)
}
    \item{prot_seq_val}{
amino acid sequence for validation (default: NULL)
}
    \item{epochs}{
number of epochs
}
    \item{batch_size}{
batch size
}
    \item{preprocessing}{
list of preprocessed results, they are set to NULL as default
x_train, y_train, lenc, length_seq, num_AA, and embedding_dim must be required for training
\itemize{
    \item x_train : embedded sequence data for train, result of the function "DeepPINCS::get_seq_encode_pad"
    \item x_val : embedded sequence data for validation, result of the function "DeepPINCS::get_seq_encode_pad"
    \item y_train : labels for train
    \item y_val : labels for validation
    \item lenc : encoded labels, result of the function "DeepPINCS::get_seq_encode_pad"
    \item length_seq : length of sequence
    \item num_AA : number of amino acids, result of the function "DeepPINCS::get_seq_encode_pad"
    \item embedding_dim : dimension of the dense embedding
    \item removed_prot_seq : index for removed protein sequences while checking
    \item removed_prot_seq_val : index for removed protein sequences of validation
}
}
    \item{use_generator}{
use data generator if TRUE (default: FALSE)
}
    \item{optimizer}{
name of optimizer (default: adam)
}
    \item{metrics}{
name of metrics (default: accuracy)
}
    \item{validation_split}{
proportion of validation data, it is ignored when there is a validation set (default: 0)
}
    \item{...}{
additional parameters for the "fit"
}
    \item{x}{
result of the function "fit_ART"
}
    \item{seed_prot}{
sequence to be used as a seed protein
}
    \item{length_AA}{
length of amino acids to be generated
}
    \item{method}{
"greedy", "beam", "temperature", "top_k", or "top_p"
}
    \item{b}{
beam size in the beam search
}
    \item{t}{
temperature in the temperature sampling (default: 1)
}
    \item{k}{
number of amino acids in the top-k sampling
}
    \item{p}{
minimum probabilty for the set of amino acids in the top-p sampling
}
}


\value{
    \item{model}{
trained ART model
}
    \item{preprocessing}{
preprocessed results
}
}


\author{
Dongmin Jung
}


\references{
Deepak, P., Chakraborty, T., & Long, C. (2021). Data Science for Fake News: Surveys and Perspectives (Vol. 42). Springer.

Liu, Z., Lin, Y., & Sun, M. (2020). Representation learning for natural language processing. Springer.

Madani, A., McCann, B., Naik, N., Keskar, N. S., Anand, N., Eguchi, R. R., Huang, P., & Socher, R. (2020). Progen: Language modeling for protein generation. arXiv:2004.03497.
}


\seealso{
keras::fit, keras::compile, ttgsea::sampling_generator, DeepPINCS::multiple_sampling_generator, DeepPINCS::seq_preprocessing, DeepPINCS::get_seq_encode_pad, CatEncoders::LabelEncoder.fit, CatEncoders::transform, CatEncoders::inverse.transform
}


\examples{
if (keras::is_keras_available() & reticulate::py_available()) {
    prot_seq <- DeepPINCS::SARS_CoV2_3CL_Protease
    
    # model parameters
    length_seq <- 10
    embedding_dim <- 16
    num_heads <- 2
    ff_dim <- 16
    num_transformer_blocks <- 2
    batch_size <- 32
    epochs <- 2
    
    # ART
    ART_result <- fit_ART(prot_seq = prot_seq,
                        length_seq = length_seq,
                        embedding_dim = embedding_dim,
                        num_heads = num_heads,
                        ff_dim = ff_dim,
                        num_transformer_blocks = num_transformer_blocks,
                        layers = list(layer_dropout(rate = 0.1),
                                    layer_dense(units = 32, activation = "relu"),
                                    layer_dropout(rate = 0.1)),
                        prot_seq_val = prot_seq,
                        epochs = epochs,
                        batch_size = batch_size,
                        use_generator = TRUE,
                        callbacks = callback_early_stopping(
                            monitor = "val_loss",
                            patience = 10,
                            restore_best_weights = TRUE))
    
    seed_prot <- "SGFRKMAFPS"
    gen_ART(ART_result, seed_prot, length_AA = 20, method = "greedy")
    gen_ART(ART_result, seed_prot, length_AA = 20, method = "beam", b = 5)
    gen_ART(ART_result, seed_prot, length_AA = 20, method = "temperature", t = 0.1)
    gen_ART(ART_result, seed_prot, length_AA = 20, method = "top_k", k = 3)
    gen_ART(ART_result, seed_prot, length_AA = 20, method = "top_p", p = 0.75)
    
    
    ### from preprocessing
    ART_result2 <- fit_ART(num_heads = 4,
                        ff_dim = 32,
                        num_transformer_blocks = 3,
                        layers = list(layer_dropout(rate=0.1),
                                    layer_dense(units=32, activation="relu"),
                                    layer_dropout(rate=0.1)),
                        epochs = epochs,
                        batch_size = batch_size,
                        preprocessing = ART_result$preprocessing,
                        use_generator = TRUE,
                        callbacks = callback_early_stopping(
                                monitor = "val_loss",
                                patience = 50,
                                restore_best_weights = TRUE))
    
    gen_ART(ART_result2, seed_prot, length_AA = 20, method = "greedy")
    gen_ART(ART_result2, seed_prot, length_AA = 20, method = "beam", b = 5)
    gen_ART(ART_result2, seed_prot, length_AA = 20, method = "temperature", t = 0.1)
    gen_ART(ART_result2, seed_prot, length_AA = 20, method = "top_k", k = 3)
    gen_ART(ART_result2, seed_prot, length_AA = 20, method = "top_p", p = 0.75)
}
}

\name{GAN}
\alias{fit_GAN}
\alias{gen_GAN}

\title{
Generative adversarial network for generating protein sequences
}

\description{
The generative adversarial network (GAN) is made up of a discriminator and a generator that compete in a two-player minimax game. The objective of the generator is to produce an output that is so close to real that it confuses the discriminator in being able to differentiate the fake data from the real data. The conditional GAN (CGAN) is based on vanilla GAN with additional conditional input to generator and discriminator. The auxiliary classifier GAN (ACGAN) is an extension of CGAN that adds conditional input only to the generator. The Word2vec is applied to amino acids for embedding. The GAN or ACGAN model can be trained by the function "fit_GAN", and then the function "gen_GAN" generates protein sequences from the trained model.
}

\usage{
fit_GAN(prot_seq,
        label = NULL,
        length_seq,
        embedding_dim,
        embedding_args = list(),
        latent_dim = NULL,
        intermediate_generator_layers,
        intermediate_discriminator_layers,
        prot_seq_val = NULL,
        label_val = NULL,
        epochs,
        batch_size,
        preprocessing = list(
            x_train = NULL,
            x_val = NULL,
            y_train = NULL,
            y_val = NULL,
            lenc = NULL,
            length_seq = NULL,
            num_seq = NULL,
            embedding_dim = NULL,
            embedding_matrix = NULL,
            removed_prot_seq = NULL,
            removed_prot_seq_val = NULL,
            latent_dim = NULL),
        optimizer = "adam",
        validation_split = 0)

gen_GAN(x,
        label = NULL,
        num_seq,
        remove_gap = TRUE)
}

\arguments{
    \item{prot_seq}{
aligned amino acid sequence
}
    \item{label}{
label (default: NULL)
}
    \item{length_seq}{
length of sequence
}
    \item{embedding_dim}{
dimension of the dense embedding
}
    \item{embedding_args}{
list of arguments for "word2vec::word2vec" but for dim, min_count and split
}
    \item{latent_dim}{
dimension of latent vector (default: NULL)
}
    \item{intermediate_generator_layers}{
list of intermediate layers for generator, without input layer
}
    \item{intermediate_discriminator_layers}{
list of intermediate layers for discriminator, without output layer
}
    \item{prot_seq_val}{
amino acid sequence for validation (default: NULL)
}
    \item{label_val}{
label for validation (default: NULL)
}
    \item{epochs}{
number of epochs
}
    \item{batch_size}{
batch size
}
    \item{preprocessing}{
list of preprocessed results, they are set to NULL as default
x_train, length_seq, num_seq, embedding_dim and embedding_matrix must be required for training
\itemize{
    \item x_train : embedded sequence data for train
    \item x_val : embedded sequence data for validation
    \item y_train : labels for train
    \item y_val : labels for validation
    \item lenc : encoded labels
    \item length_seq : length of sequence
    \item num_seq : number of sequences for train
    \item embedding_dim : dimension of the dense embedding
    \item embedding_matrix : embedding matrix
    \item removed_prot_seq : index for removed protein sequences while checking
    \item removed_prot_seq_val : index for removed protein sequences of validation
    \item latent_dim : dimension of latent vector
}
}
    \item{optimizer}{
name of optimizer (default: adam)
}
    \item{validation_split}{
proportion of validation data, it is ignored when there is a validation set (default: 0)
}
    \item{x}{
result of the function "fit_GAN"
}
    \item{num_seq}{
number of sequences to be generated
}
    \item{remove_gap}{
remove gaps from sequences (default: TRUE)
}
}


\value{
    \item{model}{
trained GAN model
}
    \item{generator}{
trained generator model
}
    \item{discriminator}{
trained discriminator model
}
    \item{preprocessing}{
preprocessed results
}
    \item{gen_seq}{
generated sequence data
}
    \item{label}{
labels for generated sequence data
}
}


\author{
Dongmin Jung
}


\seealso{
keras::train_on_batch, keras::evaluate, keras::compile, CatEncoders::LabelEncoder.fit, CatEncoders::transform, CatEncoders::inverse.transform
}


\references{
Liebowitz, J. (Ed.). (2020). Data Analytics and AI. CRC Press.

Pedrycz, W., & Chen, S. M. (Eds.). (2020). Deep Learning: Concepts and Architectures. Springer.

Suguna, S. K., Dhivya, M., & Paiva, S. (Eds.). (2021). Artificial Intelligence (AI): Recent Trends and Applications. CRC Press.

Sun, S., Mao, L., Dong, Z., & Wu, L. (2019). Multiview machine learning. Springer.
}


\examples{
if (keras::is_keras_available() & reticulate::py_available()) {
    data("example_PTEN")
    # model parameters
    length_seq <- 403
    embedding_dim <- 8
    latent_dim <- 4
    epochs <- 2
    batch_size <- 64
    
    # GAN
    GAN_result <- fit_GAN(prot_seq = example_PTEN,
                        length_seq = length_seq,
                        embedding_dim = embedding_dim,
                        latent_dim = latent_dim,
                        intermediate_generator_layers = list(
                            layer_dense(units = 16),
                            layer_dense(units = 128)),
                        intermediate_discriminator_layers = list(
                            layer_dense(units = 128, activation = "relu"),
                            layer_dense(units = 16, activation = "relu")),
                        prot_seq_val = example_PTEN,
                        epochs = epochs,
                        batch_size = batch_size)
    set.seed(1)
    gen_prot_GAN <- gen_GAN(GAN_result, num_seq = 100)
    
    
    ### from preprocessing
    GAN_result2 <- fit_GAN(preprocessing = GAN_result$preprocessing,
                            intermediate_generator_layers = list(
                                layer_dense(units = 16),
                                layer_dense(units = 128)),
                            intermediate_discriminator_layers = list(
                                layer_dense(units = 128, activation = "relu"),
                                layer_dense(units = 16, activation = "relu")),
                            epochs = epochs,
                            batch_size = batch_size)
    gen_prot_GAN <- gen_GAN(GAN_result2, num_seq = 100)
}
}

\name{VAE}
\alias{fit_VAE}
\alias{gen_VAE}

\title{
Variational autoencoder for generating protein sequences
}

\description{
The variational autoencoder (VAE) is a class of autoencoder where the encoder module is used to learn the parameter of a distribution and the decoder is used to generate examples from samples drawn from the learned distribution. The conditional variational autoencoder (CVAE) is designed to generate desired samples by including additional conditioning information. Since there may be underlying distinctions between groups of samples, the Gaussian mixture model is used for sequence generation. The Word2vec is applied to amino acids for embedding. The VAE or CVAE model can be trained by the function "fit_VAE", and then the function "gen_VAE" generates protein sequences from the trained model.
}

\usage{
fit_VAE(prot_seq,
        label = NULL,
        length_seq,
        embedding_dim,
        embedding_args = list(),
        latent_dim = 2,
        intermediate_encoder_layers,
        intermediate_decoder_layers,
        prot_seq_val = NULL,
        label_val = NULL,
        regularization = 1,
        epochs,
        batch_size,
        preprocessing = list(
            x_train = NULL,
            x_val = NULL,
            y_train = NULL,
            y_val = NULL,
            lenc = NULL,
            length_seq = NULL,
            embedding_dim = NULL,
            embedding_matrix = NULL,
            removed_prot_seq = NULL,
            removed_prot_seq_val = NULL),
        use_generator = FALSE,
        optimizer = "adam",
        validation_split = 0, ...)

gen_VAE(x,
        label = NULL,
        num_seq,
        remove_gap = TRUE,
        batch_size,
        use_generator = FALSE)
}

\arguments{
    \item{prot_seq}{
aligned amino acid sequence
}
    \item{label}{
label (default: NULL)
}
    \item{length_seq}{
length of sequence
}
    \item{embedding_dim}{
dimension of the dense embedding
}
    \item{embedding_args}{
list of arguments for "word2vec::word2vec" but for dim, min_count and split
}
    \item{latent_dim}{
dimension of latent vector (default: 2)
}
    \item{intermediate_encoder_layers}{
list of intermediate layers for encoder, without input layer
}
    \item{intermediate_decoder_layers}{
list of intermediate layers for decoder, without output layer
}
    \item{regularization}{
regularization parameter, which is nonnegative (default: 1)
}
    \item{prot_seq_val}{
amino acid sequence for validation (default: NULL)
}
    \item{label_val}{
label for validation (default: NULL)
}
    \item{epochs}{
number of epochs
}
    \item{batch_size}{
batch size
}
    \item{preprocessing}{
list of preprocessed results, they are set to NULL as default
x_train, length_seq, embedding_dim and embedding_matrix must be required for training
\itemize{
    \item x_train : embedded sequence data for train
    \item x_val : embedded sequence data for validation
    \item y_train : labels for train
    \item y_val : labels for validation
    \item lenc : encoded labels
    \item length_seq : length of sequence
    \item embedding_dim : dimension of the dense embedding
    \item embedding_matrix : embedding matrix
    \item removed_prot_seq : index for removed protein sequences while checking
    \item removed_prot_seq_val : index for removed protein sequences of validation
}
}
    \item{use_generator}{
use data generator if TRUE (default: FALSE)
}
    \item{optimizer}{
name of optimizer (default: adam)
}
    \item{validation_split}{
proportion of validation data, it is ignored when there is a validation set (default: 0)
}
    \item{...}{
additional parameters for the "fit"
}
    \item{x}{
result of the function "fit_VAE"
}
    \item{num_seq}{
number of sequences to be generated
}
    \item{remove_gap}{
remove gaps from sequences (default: TRUE)
}
}


\value{
    \item{model}{
trained VAE model
}
    \item{encoder}{
trained encoder model
}
    \item{decoder}{
trained decoder model
}
    \item{preprocessing}{
preprocessed results
}
    \item{gen_seq}{
generated sequence data
}
    \item{label}{
labels for generated sequence data
}
    \item{latent_vector}{
latent vector from embedded sequence data
}
}


\author{
Dongmin Jung
}


\references{
Cinelli, L. P., Marins, M. A., da Silva, E. A. B., & Netto, S. L. (2021). Variational Methods for Machine Learning with Applications to Deep Networks. Springer.

Liebowitz, J. (Ed.). (2020). Data Analytics and AI. CRC Press.
}


\seealso{
keras::fit, keras::compile, reticulate::array_reshape, mclust::mclustBIC, mclust::mclustModel, mclust::sim, DeepPINCS::multiple_sampling_generator, CatEncoders::LabelEncoder.fit, CatEncoders::transform, CatEncoders::inverse.transform
}


\examples{
if (keras::is_keras_available() & reticulate::py_available()) {
    data("example_luxA")
    label <- substr(example_luxA, 3, 3)
    
    # model parameters
    length_seq <- 360
    embedding_dim <- 8
    batch_size <- 128
    epochs <- 2
    
    # CVAE
    VAE_result <- fit_VAE(prot_seq = example_luxA,
                        label = label,
                        length_seq = length_seq,
                        embedding_dim = embedding_dim,
                        embedding_args = list(iter = 20),
                        intermediate_encoder_layers = list(layer_dense(units = 128),
                                                            layer_dense(units = 16)),
                        intermediate_decoder_layers = list(layer_dense(units = 16),
                                                            layer_dense(units = 128)),
                        prot_seq_val = example_luxA,
                        label_val = label,
                        epochs = epochs,
                        batch_size = batch_size,
                        use_generator = FALSE,
                        optimizer = keras::optimizer_adam(clipnorm = 0.1),
                        callbacks = keras::callback_early_stopping(
                            monitor = "val_loss",
                            patience = 10,
                            restore_best_weights = TRUE))
    gen_prot_VAE_I <- gen_VAE(VAE_result, label = rep("I", 100), num_seq = 100)
    gen_prot_VAE_L <- gen_VAE(VAE_result, label = rep("L", 100), num_seq = 100)
    
    
    ### from preprocessing
    VAE_result2 <- fit_VAE(intermediate_encoder_layers = list(layer_dense(units = 128),
                                                            layer_dense(units = 16)),
                            intermediate_decoder_layers = list(layer_dense(units = 16),
                                                            layer_dense(units = 128)),
                            epochs = epochs, batch_size = batch_size,
                            preprocessing = VAE_result$preprocessing,
                            use_generator = FALSE,
                            optimizer = keras::optimizer_adam(clipnorm = 0.1),
                            callbacks = keras::callback_early_stopping(
                                monitor = "val_loss",
                                patience = 10,
                                restore_best_weights = TRUE))
    gen_prot_VAE2_I <- gen_VAE(VAE_result2, label = rep("I", 100), num_seq = 100)
    gen_prot_VAE2_L <- gen_VAE(VAE_result2, label = rep("L", 100), num_seq = 100)
}
}

\name{prot_seq_check}
\alias{prot_seq_check}


\title{
Check a protein sequence
}

\description{
The protein sequence dataset is filtered by eliminating sequences containing the non-amino acid characters (digits and blank spaces) from the amino acid sequences. A valid amino acid sequence means a string that only contains capital letters of an alphabet and a hyphen for a gap.
}

\usage{
prot_seq_check(prot_seq, label = NULL)
}

\arguments{
    \item{prot_seq}{
amino acid sequences
}
    \item{label}{
label (default: NULL)
}
}

\value{
valid sequences
}

\author{
Dongmin Jung
}

\references{
Mukhopadhyay, C. S., Choudhary, R. K., & Iquebal, M. A. (2017). Basic Applied Bioinformatics. John Wiley & Sons.
}

\examples{
data("example_PTEN")
prot_seq_check(example_PTEN[1])
}

\name{example_luxA}
\alias{example_luxA}

\title{
Example Data for Protein Sequences
}

\description{
The data consist of selected amino acid sequences of the luxA. There are 2283 aligned sequences of length 360.
}

\usage{
example_luxA
}

\value{
aligned amino acid sequences
}

\source{
Hawkins-Hooker, A., Depardieu, F., Baur, S., Couairon, G., Chen, A., & Bikard, D. (2020). Generating functional protein variants with variational autoencoders. bioRxiv.
}

\author{
Dongmin Jung
}
